import importlib
import sys
from unittest import mock

from django.conf import settings
from django.contrib.auth.models import User
from django.urls import reverse

from otree.common import (
    EMPTY_ADMIN_USERNAME_MSG,
    EMPTY_ADMIN_PASSWORD_MSG,
    ensure_superuser_exists,
)

# from channels.tests import ChannelLiveServerTestCase
from .utils import TestCase, get_path


def reload_urlconf():
    if settings.ROOT_URLCONF in sys.modules:
        module = sys.modules[settings.ROOT_URLCONF]
    else:
        module = importlib.import_module(settings.ROOT_URLCONF)
    urlpatterns = module.urlpatterns
    urlpatterns.clear()
    urlpatterns += module.get_urlpatterns()


class TestAdminBasic(TestCase):
    def test_load_pages(self):
        self.client.login()
        for tab in ['demo', 'sessions', 'rooms', 'create_session']: # , 'server_check' # makes a network request
            response = self.client.get('/{}/'.format(tab), follow=True)
            self.assertEqual(response.status_code, 200)

    def test_login(self):
        login_url = '/accounts/login/'
        resp = self.client.post(
            login_url,
            data={
                'username': settings.ADMIN_USERNAME,
                'password': settings.ADMIN_PASSWORD,
            },
            follow=True,
        )

        self.assertEqual(resp.status_code, 200)
        self.assertNotIn(login_url, get_path(resp, if_no_redirect=login_url))

    def test_auth_level(self):

        # whether you should be granted access when not logged in,
        # based on AUTH_LEVEL
        auth_level_outcomes = {
            'STUDY': {'DemoIndex': False, 'ExportIndex': False},
            'DEMO': {'DemoIndex': True, 'ExportIndex': False},
            '': {'DemoIndex': True, 'ExportIndex': True},
        }

        for AUTH_LEVEL in auth_level_outcomes:

            with self.settings(AUTH_LEVEL=AUTH_LEVEL):
                reload_urlconf()
                for url_name in auth_level_outcomes[AUTH_LEVEL]:
                    url = reverse(url_name)
                    resp = self.client.get(url, follow=True)
                    # resp.render()
                    assert settings.AUTH_LEVEL == AUTH_LEVEL
                    is_login_page = any(
                        ['accounts/login' in ele[0] for ele in resp.redirect_chain]
                    )
                    access_granted = not is_login_page
                    self.assertEqual(
                        access_granted,
                        auth_level_outcomes[AUTH_LEVEL][url_name],
                        msg='AUTH_LEVEL={}, url_name={}, access_granted={}'.format(
                            AUTH_LEVEL, url_name, access_granted
                        ),
                    )
        # reload URLconf to restore back to its original state,
        # so other tests can run normally
        reload_urlconf()


class SuperuserTests(TestCase):
    def test_login_page(self):

        br = self.splinter_browser
        failure_msg = 'problemo'
        login_url = '/accounts/login/'
        with mock.patch(
            'otree.common.ensure_superuser_exists', return_value=failure_msg
        ):
            br.visit(login_url)
            assert br.find_by_text(failure_msg)
        br.fill('username', 'admin')
        br.fill('password', 'wrongpassword')
        button = br.find_by_id('btn-login')
        button.click()

        # it should tell them that resetting the database may be necessary
        self.assertIn('reset the database', br.html)

    def test_superuser_creation(self):
        # don't use 'admin' because that may have already been created
        # when launching tests
        username = 'admin2'

        with self.settings(ADMIN_USERNAME='', ADMIN_PASSWORD='aoueaeu'):
            self.assertEqual(ensure_superuser_exists(), EMPTY_ADMIN_USERNAME_MSG)

        with self.settings(ADMIN_USERNAME=username, ADMIN_PASSWORD=''):
            self.assertEqual(ensure_superuser_exists(), EMPTY_ADMIN_PASSWORD_MSG)
        self.assertFalse(User.objects.filter(username=username).exists())

        with self.settings(ADMIN_USERNAME='admin2', ADMIN_PASSWORD='aoueaeu'):
            self.assertEqual(ensure_superuser_exists(), '')
        self.assertTrue(User.objects.filter(username=username).exists())


# We can't enable these test cases until we upgrade to CHannels 1.0+
'''
class TestAdminJS(ChannelLiveServerTestCase):

    def setUp(self):
        self.browser = splinter.Browser('phantomjs')

    def test_create_session(self):
        br = self.browser
        create_session_url = '{}{}'.format(self.live_server_url, reverse('CreateSession'))
        br.visit(create_session_url)
        br.fill_form({
            'session_config': 'simple',
            'num_participants': '1',
        })
        button = br.find_by_value('Create')
        button.click()
        self.assertTrue(br.is_text_present('Simple Game: session', wait_time=3))
'''
