import pytest
from otree.models.session import Session
from otree.bots.bot import ParticipantBot
from .utils import get_ws_communicator
from .utils import TestCase, OTreeTestClient, execute_mocked_send_group
from otree.session import create_session
from django.urls import reverse
import otree.channels.utils as channel_utils
from otree import common
from unittest.mock import patch


# use a wrapper so that unittest doesn't find this base class at the
# module level, otherwise it will run it.
class WrapperToHideClass:
    class BaseTestCase(TestCase):
        def advance_to_end(self):

            session = self.session
            participants = session.get_participants()
            max_page_index = participants[0]._max_page_index

            # once to open the start links, then num_pages to get to last page
            for x in range(max_page_index + 1):
                session.advance_last_place_participants()

            participants = session.get_participants()
            for p in participants:
                self.assertGreaterEqual(p._index_in_pages, max_page_index)

        def test_advance_to_end(self):
            self.advance_to_end()

        def test_some_already_started(self):
            p = self.session.participant_set.first()

            bot = ParticipantBot(p, player_bots=[])
            bot.open_start_url()
            self.advance_to_end()


class TestAdvanceSlowest(WrapperToHideClass.BaseTestCase):
    def setUp(self):
        create_session('advance_slowest', num_participants=2)
        self.session = Session.objects.get()


class TestAdvanceSlowestWaitPageFirst(WrapperToHideClass.BaseTestCase):
    def setUp(self):
        create_session('advance_slowest_wait', num_participants=2)
        self.session = Session.objects.get()


@pytest.mark.django_db(transaction=True)
@pytest.mark.asyncio
async def test_views():
    # need to use context manager because it seems mock decorators
    # don't work on async functions
    with patch('otree.channels.utils._sync_group_send') as mock_a2s:
        client = OTreeTestClient()
        client.login()
        session = create_session('two_rounds_1p', num_participants=1)
        p_code = session.participant_set.values_list('code', flat=True).first()

        start_url = common.participant_start_url(p_code)
        client.get(start_url, follow=True)

        # can't use splinter Django browser because "advance users" uses AJAX
        advance_url = reverse('AdvanceSession', args=[session.code])
        client.post(advance_url)
        await execute_mocked_send_group(mock_a2s)

        # CASE 1: slow websocket, connects after auto-advance
        ws_client = get_ws_communicator(
            channel_utils.auto_advance_path(participant_code=p_code, page_index=1)
        )
        await ws_client.connect(timeout=4)

        assert await ws_client.receive_json_from() == {'auto_advanced': True}

        await ws_client.disconnect()

        # easy way to get to 2nd page
        client.get(start_url, follow=True)

        # CASE 2: websocket connects before auto-advance
        ws_client = get_ws_communicator(
            path=channel_utils.auto_advance_path(participant_code=p_code, page_index=2)
        )
        await ws_client.connect(timeout=4)

        client.post(advance_url)
        await execute_mocked_send_group(mock_a2s)

        assert await ws_client.receive_json_from() == {'auto_advanced': True}
