import logging
import unittest.mock
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pytest

import otree.session
import tests.bots_cases.pages
from otree.api import Submission, SubmissionMustFail
from otree.api import expect, Currency
from otree.bots.bot import (
    MissingHtmlButtonError,
    MissingHtmlFormFieldError,
    ParticipantBot,
    ExpectError,
)
from otree.bots.runner import run_bots, make_bots
from otree.common import BotError
from tests.bots_cases.tests import PlayerBot
from .utils import TestCase


class TestBots(TestCase):
    def test_bot_runs(self):

        session = otree.session.create_session(
            session_config_name='bots_raise', num_participants=1
        )

        with self.assertRaises(ZeroDivisionError):
            run_bots(session)

    def test_bots_check_html(self):

        session = otree.session.create_session(
            session_config_name='bots_check_html', num_participants=3
        )

        b1, b2, b3 = make_bots(
            session_pk=session.pk, case_number=0, use_browser_bots=False
        )

        with self.settings(BOTS_CHECK_HTML=True):
            with self.assertRaises(MissingHtmlFormFieldError):
                b1._play_individually()

            with self.assertRaises(MissingHtmlButtonError):
                b2._play_individually()

            # should run without problems
            b3._play_individually()

    def test_bot_bad_post(self):
        """
        Test that posting bad data without using SubmitInvalid
        will raise an error.
        """

        session = otree.session.create_session(
            session_config_name='bots_bad_post', num_participants=1
        )

        # need to disable log output, because this triggers an exception
        # that is logged to stdout
        logging.disable(logging.CRITICAL)

        with self.assertRaises(BotError):
            run_bots(session)

        logging.disable(logging.NOTSET)

    def test_bots_submission_varieties(self):
        '''Testing different syntaxes for a submission'''
        Page1 = tests.bots_cases.pages.Page1
        for submission in [
            Submission(Page1),
            Submission(Page1, {}),
            SubmissionMustFail(Page1, check_html=False, error_fields=['f2']),
            Submission(Page1, {'f2': True}, check_html=False),
            SubmissionMustFail(Page1, {}),
            Submission(Page1, {'f3': True}),
        ]:
            self.assertIsInstance(submission, dict)

    @unittest.mock.patch.object(PlayerBot, 'case1')
    @unittest.mock.patch.object(PlayerBot, 'case2')
    def test_cases(self, patched_case2, patched_case1):
        '''
        Test that all cases are run
        '''
        from otree.bots.runner import run_all_bots_for_session_config

        run_all_bots_for_session_config('bots_cases', 1, export_path=None)

        patched_case1.assert_called()
        patched_case2.assert_called()

    def test_form_errors(self):
        session = otree.session.create_session(
            session_config_name='bots_form_errors', num_participants=1
        )

        run_bots(session, case_number=0)

    @patch('otree.bots.runner.run_bots')
    def test_export(self, *args):
        from otree.bots.runner import run_all_bots_for_session_config

        with TemporaryDirectory() as tempdir:
            run_all_bots_for_session_config('bots_simple', 3, export_path=tempdir)
            csv_text = Path(tempdir, 'tests.bots_simple.csv').read_text('utf8')
            numlines = len([l for l in csv_text.split('\n') if l.strip()])
            self.assertEqual(numlines, 4)  # header + 3 lines


class TestExpectFunction:
    def test_bad(self):
        with pytest.raises(ExpectError) as excinfo:
            expect('a', 'b')
        assert str(excinfo.value) == "Expected 'b', actual value is 'a'"

        with pytest.raises(ExpectError) as excinfo:
            expect(Currency(0), '>', 0)
        assert str(excinfo.value) == 'Assertion failed: Currency(0.00) > 0'

        with pytest.raises(ExpectError) as excinfo:
            expect(0, 1.5)
        assert str(excinfo.value) == 'Expected 1.5, actual value is 0'

        with pytest.raises(ExpectError) as excinfo:
            expect('you won', 'in', '<html>you lost</html>')

    def test_good(self):

        expect('a', 'in', 'ab')
        expect(0, '<=', 0)
        expect(Currency(0), Currency(0))

    def test_bad_args(self):
        with pytest.raises(ValueError):
            expect(1, 'blah', '2')
