from json import dumps
import logging
from typing import List
from unittest.mock import MagicMock, patch
from collections import defaultdict
from queue import Queue
import os
from django.contrib.auth.models import User
from django.core.management import call_command
from django.shortcuts import reverse

import otree.bots.browser as browser_bots
import otree.channels.utils as channel_utils
from otree import common
from otree.models import Participant
from otree.session import create_session
from tests.bots_cases.tests import PlayerBot
from .utils import (
    TestCase,
    get_ws_communicator,
    AsyncTestCase,
    async_test,
    execute_mocked_send_group,
)

from .utils import REST_KEY


class CreateSessionTests(AsyncTestCase):
    def setUp(self):
        super().setUp()
        self.cl2 = self.client_class(HTTP_OTREE_REST_KEY=REST_KEY)

    @patch.dict(os.environ, {'OTREE_REST_KEY': REST_KEY})
    @patch('otree.channels.utils._sync_group_send')
    @patch.object(PlayerBot, '__init__', return_value=None)
    @async_test
    async def test_cases(self, patched_init, mocked):
        browser_bots.browser_bot_worker = browser_bots.BotWorker()

        pre_create_id = 'bogus_pcid'
        payload = dict(
            session_config_name='bots_cases',
            num_participants=1,
            case_number=1,
            pre_create_id=pre_create_id,
        )
        resp = self.cl2.post(
            reverse('CreateBrowserBotsSession'),
            dumps(payload),
            content_type="application/json",
        )
        await execute_mocked_send_group(mocked)
        calls = patched_init.call_args_list
        self.assertEqual(len(calls), 1)
        for call in calls:
            kwargs = call[1]
            self.assertEqual(kwargs['case_number'], 1)
        # now access the start link as a participant
        resp = self.client.get(
            reverse('BrowserBotStartLink', args=[pre_create_id]), follow=False
        )
        participant = Participant.objects.get()
        self.assertEqual(participant._start_url(), resp['Location'])

    # see test_create_session for coverage of browser bots demo session


class BrowserSimulateTests(AsyncTestCase):
    @patch('otree.channels.utils._sync_group_send')
    @async_test
    async def test_simulate(self, mocked):
        session = create_session('gbat', num_participants=6)
        await self.simulate_browser_bots(session, mocked)

    async def simulate_browser_bots(self, session, mocked_send_group):
        browser_bots.browser_bot_worker = browser_bots.BotWorker()
        worker = browser_bots.browser_bot_worker
        worker.initialize_session(session_pk=session.pk, case_number=0)

        bots = worker.browser_bots.copy()

        for bot in bots.values():
            bot.open_start_url()

        path = channel_utils.browser_bots_launcher_path(session.code)
        launcher_ws_client = get_ws_communicator(path=path)
        launcher_ws_client.scope['user'] = User.objects.first()
        await launcher_ws_client.connect()

        loops_without_progress = 0

        while True:
            if len(bots) == 0:
                break
            # bots got stuck if there's 2 wait pages in a row
            if loops_without_progress > 5:
                raise AssertionError('Bots got stuck')
            # store in a separate list so we don't mutate the iterable
            playable_ids = list(bots.keys())
            progress_made = False
            for code in playable_ids:
                bot = bots[code]
                if bot.on_wait_page():
                    pass
                elif 'browser-bot-auto-submit' in bot.response.content.decode('utf-8'):

                    bot.response = bot._client.post(bot.url, follow=True)
                    progress_made = True
                else:
                    bots.pop(code)
            if not progress_made:
                loops_without_progress += 1

        unseen_participant_codes = session.participant_set.values_list(
            'code', flat=True
        )
        unseen_participant_codes = set(unseen_participant_codes)

        await execute_mocked_send_group(mocked_send_group)
        for i in range(session.num_participants):
            participant_code = await launcher_ws_client.receive_from()
            self.assertNotEqual(participant_code, None)
            unseen_participant_codes.remove(participant_code)
        self.assertEqual(len(unseen_participant_codes), 0)


class FakeRequestsSession:
    cookies = {'csrftoken': 'FAKE_CSRF_TOKEN'}
    headers = {}

    def post(self, url, json, **kwargs):
        return MagicMock(ok=True, status_code=200, url='/')


class FakeSessionConfig(dict):
    def get_num_bot_cases(self):
        return 1


class LauncherTests(TestCase):
    @patch('otree.bots.browser_launcher.Popen', autospec=True)
    @patch('otree.bots.browser_launcher.check_output', autospec=True)
    @patch(
        'otree.bots.browser_launcher.run_websocket_client_until_finished', autospec=True
    )
    @patch('otree.bots.browser_launcher.requests_session', FakeRequestsSession)
    def test_happy_path(
        self,
        fake_run_websocket: MagicMock,
        fake_check_output: MagicMock,
        fake_popen: MagicMock,
    ):
        fake_run_websocket.return_value = 1.0
        fake_check_output.return_value = b'fake ps output'

        FAKE_SESSION_NAME = 'fake_session_name'

        with patch(
            'otree.bots.browser_launcher.SESSION_CONFIGS_DICT',
            {
                FAKE_SESSION_NAME: FakeSessionConfig(
                    name=FAKE_SESSION_NAME, num_demo_participants=1
                )
            },
        ):
            call_command('browser_bots', FAKE_SESSION_NAME, 1)
            fake_popen.assert_called()

    def test_start_url_no_session(self):
        # above there is a test for opening the link if a session already exists
        self.client.login()
        resp = self.client.get(
            reverse('BrowserBotStartLink', args=['not_yet_created_id'])
        )
        self.assertTemplateUsed(resp, template_name='otree/WaitPage.html')
