from django.core.signing import Signer
from django.template import Context, Template
from django.test import LiveServerTestCase
from django.urls import reverse
import pytest

import otree.channels.utils as channel_utils
from otree.chat import ChatTagError
from otree.models_concrete import ChatMessage
from otree.session import create_session
from tests.chatapp.models import Player, Constants
from .utils import OTreeChromeBrowser
from .utils import TestCase, AsyncTestCase, async_test, get_ws_communicator


class TagTests(TestCase):
    def test_chat(self):
        session = create_session('chat', num_participants=1)
        player = Player.objects.filter(session=session).first()

        self.context = Context(
            {
                'participant': player.participant,
                'session': session,
                'player': player,
                'group': player.group,
                'subsession': player.subsession,
                'Constants': Constants,
            }
        )

        for fragment in [
            '{% chat %}',
            '{% chat channel="abc" nickname="def" %}',
            '''{% chat nickname="中华人民共和国 '-/;," %}''',
            '{% chat nickname=123 channel=456 %}',
        ]:
            # for now, just make sure it doesn't crash
            self.render(fragment)

        for fragment in [
            '{% chat channel="contains spaces" %}',
            """{% chat channel="contains 'quotes'" %}""" '{% chat channel="中华" %}',
        ]:
            with self.assertRaises(ChatTagError):
                self.render(fragment)

    def render(self, fragment):
        template = Template('{% load otree %}' + fragment)
        return template.render(self.context)




class ConsumerTests(AsyncTestCase):
    @async_test
    async def test_connect_before_message(self):
        session = create_session('chat', num_participants=2)
        p1, p2 = session.get_participants()

        channel_name = 'random-channel-name'

        path1 = channel_utils.chat_path(channel_name, p1.id)
        path2 = channel_utils.chat_path(channel_name, p2.id)

        c1 = get_ws_communicator(path1)
        c2 = get_ws_communicator(path2)

        for c in [c1, c2]:
            connected, _ = await c.connect()
            assert connected
            await c.receive_json_from()
        # insights:
        # - need to use receive_json_from, which actually pops something
        # from the output queue, versus receive_nothing, which just checks
        # whether it's empty.
        # - need to call receive_json_from exactly the same number of times
        # as the numbers of messages that are sent.

        await c1.send_json_to({'body': 'hello', 'nickname_signed': Signer().sign('p1')})
        # if I add this line, then post_receive_json is executed
        # await c1.receive_from()

        expected_json = [{'nickname': 'p1', 'body': 'hello', 'participant_id': p1.id}]

        assert await c1.receive_json_from() == expected_json
        assert await c2.receive_json_from() == expected_json

    @async_test
    async def test_message_before_connect(self):
        session = create_session('chat', num_participants=2)
        p1, p2 = session.get_participants()

        chat_room_name = 'random-chat-room-name'

        path1 = channel_utils.chat_path(chat_room_name, p1.id)
        path2 = channel_utils.chat_path(chat_room_name, p2.id)

        c1 = get_ws_communicator(path1)
        c2 = get_ws_communicator(path2)

        connected, _ = await c1.connect()
        assert connected
        await c1.receive_json_from()

        for body in ['body1', '中华']:
            data = {'body': body, 'nickname_signed': Signer().sign('p1')}
            await c1.send_json_to(data)
            assert (await c1.receive_json_from())[0]['body'] == data['body']

        await c2.connect()

        self.assertEqual(
            await c2.receive_json_from(),
            [
                {'nickname': 'p1', 'body': 'body1', 'participant_id': p1.id},
                {'nickname': 'p1', 'body': '中华', 'participant_id': p1.id},
            ],
        )


@pytest.mark.slow
class BrowserTests(LiveServerTestCase):
    def test_browser(self):
        session = create_session('chat', num_participants=1)

        br = OTreeChromeBrowser(self.live_server_url)
        participant = session.participant_set.first()
        start_url = participant._start_url()
        br.go(start_url)

        self.assertEqual(br.evaluate_script('chatTestsPassed'), True)


class ExportTests(TestCase):
    def setUp(self):
        self.splinter_browser.login()

    def test_export_nodata(self):
        session = create_session('chat', num_participants=1)

        # even if a session exists, what matters is whether there are chat
        # messages
        br = self.splinter_browser
        br.visit_reverse('ExportIndex')
        found_links = br.find_link_by_partial_href('otreechatcore_export')
        self.assertEqual(len(found_links), 0)

    def test_export(self):
        session = create_session('chat', num_participants=1)
        participant = session.participant_set.first()

        ChatMessage.objects.create(
            channel='bogus-channel',
            participant=participant,
            nickname='Nick Name',
            body='Testing 123',
        )

        ChatMessage.objects.create(
            channel='bogus-channel-2',
            participant=participant,
            nickname='Nick Name',
            body='Testing 456',
        )

        br = self.splinter_browser
        br.visit_reverse('ExportIndex')
        # make sure the link is there, but can't download it with Splinter
        found_links = br.find_link_by_partial_href('otreechatcore_export')
        self.assertEqual(len(found_links), 1)

        response = self.client.get(reverse('ExportChat'))

        content_disposition = response["Content-Disposition"]
        content_type = response["content-type"]

        self.assertEqual(content_type, "text/csv")

        self.assertTrue(content_disposition.startswith('attachment; filename='))

        csv_text = response.content.decode('utf-8')

        rows = [row for row in csv_text.split('\n') if row.strip()]

        # 2 chat records + header row
        self.assertEqual(len(rows), 3)

        header_row = rows[0]

        for expected_text in ['participant__code', 'timestamp', 'body']:
            self.assertIn(expected_text, header_row)
