import os
from unittest import mock
from unittest.mock import patch

from django.conf import LazySettings
from django.core.management import call_command

import otree_startup
from otree.management.commands.bots import MSG_BOTS_HELP
from otree.management.commands.startproject import Command as StartProjectCommand
from otree.management.commands.prodserver1of2 import get_addr_port
from .utils import TestCase


class ExecFromCLITests(TestCase):
    @mock.patch("otree_startup.django_settings", new=LazySettings())
    @mock.patch("otree_startup.__version__", "foo")
    @mock.patch("sys.stdout")
    @mock.patch('otree_startup.argv', "otree version".split())
    def test_execute_from_command_line_no_env_command(self, stdout):
        otree_startup.execute_from_command_line()
        stdout.write.assert_called_with("foo\n")

    # execute_from_command_line() ignores its arguments, so need to patch argv
    @mock.patch("otree_startup.argv", new="otree startproject foo".split())
    @mock.patch("otree_startup.django_settings", new=LazySettings())
    @mock.patch.dict(os.environ, {'DJANGO_SETTINGS_MODULE': 'settings_NOTFOUND'})
    @mock.patch.object(StartProjectCommand, 'run_from_argv')
    def test_startproject(self, startproject_run):
        otree_startup.execute_from_command_line()
        startproject_run.assert_called()

    # we want to simulate having no settings
    # order is important. LazySettings looks at
    # DJANGO_SETTINGS_MODULE, which is set by a decorator
    @mock.patch("otree_startup.argv", new="otree check".split())
    @mock.patch("otree_startup.django_settings", new=LazySettings())
    @mock.patch.dict(os.environ, {'DJANGO_SETTINGS_MODULE': 'settings_NOTFOUND'})
    def test_settings_import_error(self):
        with self.assertLogs('otree_startup') as cm:
            otree_startup.execute_from_command_line()
            self.assertIn('Cannot find oTree settings', cm.output[0])

    @mock.patch("otree_startup.argv", new="otree --version".split())
    @mock.patch("otree_startup.django_settings", new=LazySettings())
    @mock.patch("os.getcwd", return_value="foo_cwd")
    def test_add_pwd(self, *args):
        with mock.patch("sys.path", new=[]) as path:
            otree_startup.execute_from_command_line()
            self.assertEqual(path, ["foo_cwd"])

    # otherwise we get 'settings already configured' error
    @mock.patch('otree_startup.configure_settings')
    @mock.patch("otree_startup.argv", new="otree help test".split())
    def test_help_subcommand(self, *args):
        '''we can also use this to get more coverage of fetch_command'''
        with patch('sys.stdout.write') as stdout:
            otree_startup.execute_from_command_line()
            args, kwargs = stdout.call_args
            self.assertIn(MSG_BOTS_HELP, args[0])

    # otherwise we get 'settings already configured' error
    @mock.patch('otree_startup.configure_settings')
    @mock.patch("otree_startup.argv", new="otree help".split())
    def test_help(self, *args):
        '''
        there was a bug where oTree commands were missing because
        real settings were not configured.
        we can also use this to get more coverage of fetch_command
        '''
        with patch('sys.stdout.write') as stdout:
            otree_startup.execute_from_command_line()
            args, kwargs = stdout.call_args
            output = args[0]
            self.assertIn('zipserver', output)
            self.assertIn('devserver', output)


class CallCommandTests(TestCase):
    @mock.patch.dict('os.environ', {'PORT': '5000'})
    def test_parse_addrport(self):
        specs = [
            # new syntax, to be consistent with runserver
            ('127.0.0.1:80', '127.0.0.1', '80'),
            # should respect the $PORT env var, for Heroku
            ('', '0.0.0.0', '5000'),
            ('8002', '0.0.0.0', '8002'),
        ]
        for addrport, expected_address, expected_port in specs:
            assert tuple(get_addr_port(addrport)) == (expected_address, expected_port)


class LegacyTests(TestCase):
    def test_legacy_augment_settings(self):
        from otree.settings import augment_settings

        initial_settings = {}
        augment_settings(initial_settings)
        # augment_settings should be a no-op.
        self.assertEqual(initial_settings, {})
