import uuid
from unittest.mock import patch

from django.contrib.auth.models import User
from django.core.management import call_command
from django.test.utils import override_settings

import otree.channels.utils as channel_utils
from otree.models import Session
from otree.session import create_session, SESSION_CONFIGS_DICT
from .saving import models as sgc_models
from .simple import models as sg_models
from .utils import TestCase, get_ws_communicator, AsyncTestCase, async_test


class TestCreateSessionsCommand(TestCase):
    def test_create_two_sessions_output(self):
        num_sessions = 2
        for i in range(num_sessions):
            call_command('create_session', 'simple', "1")
        created_sessions = Session.objects.count()
        self.assertEqual(created_sessions, num_sessions)

    def test_create_one_session(self):
        call_command('create_session', 'simple', "1")
        self.assertEqual(Session.objects.count(), 1)
        session = Session.objects.get()
        self.assertEqual(session.config['name'], 'simple')

        self.assertEqual(sg_models.Subsession.objects.count(), 1)
        subsession = sg_models.Subsession.objects.get()

        self.assertEqual(sg_models.Player.objects.count(), 1)

        player = sg_models.Player.objects.get()
        self.assertEqual(player.session, session)
        self.assertEqual(player.subsession, subsession)

    def test_session_vars(self):
        key = str(uuid.uuid4())
        value = str(uuid.uuid4())

        call_command('create_session', 'two_simple_games', "1")

        self.assertEqual(Session.objects.count(), 1)
        session = Session.objects.get()
        self.assertEqual(session.config['name'], 'two_simple_games')

        self.assertEqual(sg_models.Subsession.objects.count(), 1)
        self.assertEqual(sgc_models.Subsession.objects.count(), 1)
        subsession0 = sg_models.Subsession.objects.get()
        subsession1 = sgc_models.Subsession.objects.get()

        self.assertEqual(sg_models.Player.objects.count(), 1)
        self.assertEqual(sgc_models.Player.objects.count(), 1)

        # retrieve player of first subsession
        player0 = sg_models.Player.objects.get()

        # add a random key value
        player0.participant.vars[key] = value
        player0.participant.save()

        # retrieve player of second subsession
        player1 = sgc_models.Player.objects.get()

        # validate all
        self.assertTrue(player0.session == player1.session == session)
        self.assertEqual(player0.subsession, subsession0)
        self.assertEqual(player1.subsession, subsession1)

        # test the random key value in second subsession
        self.assertEqual(player1.participant.vars.get(key), value)

    def test_edit_session_config(self):
        '''maybe no longer needed now that we test the whole view'''
        session_config_name = 'simple'
        config_key = 'use_browser_bots'
        session_config = SESSION_CONFIGS_DICT[session_config_name]
        original_config_value = session_config[config_key]
        new_config_value = not original_config_value

        session = create_session(
            'simple',
            num_participants=1,
            modified_session_config_fields={config_key: new_config_value},
        )
        self.assertEqual(session.config[config_key], new_config_value)

        # make sure it didn't affect the value for future sessions
        # (this bug occurred because we mutated the dictionary)
        session2 = create_session('simple', num_participants=1)
        self.assertEqual(session2.config[config_key], original_config_value)


class CreateDemoSessionTests(AsyncTestCase):
    @override_settings(AUTH_LEVEL='DEMO')
    @async_test
    async def test_create(self):
        ws_client = get_ws_communicator(path=channel_utils.create_demo_session_path())
        await ws_client.connect()
        await ws_client.send_json_to(dict(session_config='simple'))

        message_dict = await ws_client.receive_json_from()
        self.assertIn('SessionStartLinks', message_dict['session_url'])

    def test_load_page(self):
        '''
        dont need to test overriding AUTH_LEVEL because that is already
        covered by another test case
        '''
        br = self.splinter_browser
        br.login()
        br.visit_reverse('CreateDemoSession', args=['simple'])
        assert br.status_code == 200


class ConsumerTests(AsyncTestCase):
    def setUp(self):
        super().setUp()
        self.splinter_browser.login()
        self.ws_client = get_ws_communicator(path=channel_utils.create_session_path())
        self.ws_client.scope['user'] = User.objects.first()

    @patch('otree.bots.browser.initialize_session')
    @async_test
    async def test_edit_config(self, patched_initialize_bots):
        br = self.splinter_browser

        config_name = 'edit_session_config'

        new_values = {
            'int': 1,
            'float': 1.57,
            'str': 'hello2',
            'participation_fee': 1,
            'use_browser_bots': True,
        }

        br.visit_reverse('CreateSession')
        form_values = {'session_config': 'edit_session_config', 'num_participants': '1'}

        for k, v in new_values.items():
            field_name = '{}.{}'.format(config_name, k)
            form_values[field_name] = str(v)

        br.fill_form(form_values)

        ws_client = self.ws_client
        await ws_client.connect()

        await ws_client.send_json_to(form_values)
        message_dict = await ws_client.receive_json_from()
        self.assertIn('SessionStartLinks', message_dict['session_url'])

        # make sure undesired fields are not present? Like you can't edit
        # 'app_sequence' or 'num_demo_participants'.
        # test fail to create session
        session = Session.objects.first()
        config = session.config

        original_config = SESSION_CONFIGS_DICT[config_name]
        for k, v in new_values.items():
            # make sure equal to new value
            self.assertEqual(config[k], v, msg=k)
            # make sure we actually changed it, and that we didn't mutate
            # the original config
            self.assertNotEqual(config[k], original_config[k], msg=k)

        patched_initialize_bots.assert_called()

    # I can't figure out how to allow it to raise but also get the message that was sent
    # before it raised
    # anyway it's OK, this is quite simple and high visibility
    @patch('otree.session.create_session', side_effect=ZeroDivisionError)
    @async_test
    async def test_traceback(self, patched):
        ws_client = self.ws_client
        await ws_client.connect()
        await ws_client.send_json_to(dict(session_config='misc_3p', num_participants=3))
        message_dict = await ws_client.receive_json_from()
        self.assertIn('error', message_dict)
        self.assertIn('traceback', message_dict)

    @async_test
    async def test_validation_error(self):
        ws_client = self.ws_client
        await ws_client.connect()
        await ws_client.send_json_to(dict(session_config='misc_3p', num_participants=4))

        message_dict = await ws_client.receive_json_from()
        self.assertNotIn('traceback', message_dict)
        self.assertIn('validation_errors', message_dict)


'''
Not working because splinter doesn't seem to recognize formaction, so I get:
Method Not Allowed (POST): /sessions/

Anyway, this test doesn't seem very useful. The test is longer than the code itself.

ToggleArchived will have the same issue

class DeleteTests(TestCase):
    def test_delete(self):

        session = create_session('simple', num_participants=1)
        br = splinter.Browser('django')
        sessions_url = reverse('Sessions')
        br.visit(sessions_url)
        # it seems splinter Django browser doesn't let you select checkboxes by value?
        # so i can't specify the session code
        #checkboxes = br.find_by_name('session')
        #for checkbox in checkboxes:
        #    checkbox.check()
        br.check('session')
        br.find_by_id('action-delete-confirm').click()
        self.assertFalse(Session.objects.filter(code=session.code).exists())
'''
