from django.shortcuts import reverse
from django.contrib.auth.models import User
from django.conf import settings
from django.core.management import call_command
from six import StringIO
from django.test import override_settings
import otree.export
import otree.session
from .utils import TestCase, AsyncTestCase, async_test, get_ws_communicator
import otree.channels.utils as channel_utils
from tests.export.models import Constants
import re
from otree.common2 import (
    make_page_completion_row,
    page_completion_buffer,
    PageTimeBatch,
    BUFFER_SIZE,
)

# Now that it's async, need to test it differently,
# using channels.


class TestExport(AsyncTestCase):

    '''
            # make sure it's found on page

    '''

    @async_test
    async def test_alignment(self):

        session_config_name = 'export'
        num_participants = 3
        num_sessions = 2

        # 3 sessions, 3 participants each is good to test alignment
        for i in range(num_sessions):
            otree.session.create_session(
                session_config_name=session_config_name,
                num_participants=num_participants,
            )

        br = self.splinter_browser
        br.login()
        br.visit_reverse('ExportIndex')
        print(br.html)
        assert br.is_element_present_by_id('app-tests.export-excel')

        c = get_ws_communicator('/export/', login=True)
        await c.connect()
        await c.send_json_to({'for_excel': '', 'app_name': 'tests.export'})
        response_dict = await c.receive_json_from()

        csv_text = response_dict['data']

        rows = [row for row in csv_text.split('\n') if row.strip()]

        # 1 row for each player + header row
        self.assertEqual(
            len(rows), num_sessions * num_participants * Constants.num_rounds + 1
        )

        header_row = rows[0]

        for expected_text in [
            'participant.id_in_session',
            'player.id_in_group',
            'player.payoff',
            'group.group_field',
            'subsession.subsession_field',
            'subsession.round_number',
            'session.code',
        ]:
            self.assertIn(expected_text, header_row)

        # 3.14 should be in the CSV without any currency formatting
        for expected_text in ['should be in export CSV', ',3.14,']:
            self.assertIn(expected_text, csv_text)

        # True/False should be converted to 1/0
        self.assertNotIn('False', csv_text)

        # test alignment
        for row in rows[1:]:
            participant_code = re.search(r',ALIGN_TO_PARTICIPANT_(\w+),', row).group(1)
            self.assertIn(',{},'.format(participant_code), row)

            session_code = re.search(r',ALIGN_TO_SESSION_(\w+),', row).group(1)
            self.assertIn(',{},'.format(session_code), row)

            group_id = re.search(r',GROUP_(\d+),', row).group(1)
            self.assertIn(',ALIGN_TO_GROUP_{},'.format(group_id), row)

            subsession_id = re.search(r',SUBSESSION_(\d+),', row).group(1)
            self.assertIn(',ALIGN_TO_SUBSESSION_{},'.format(subsession_id), row)


class TestWideCSV(TestCase):
    def test_export(self):

        session_config_name = 'export'
        app_name = 'tests.export'
        num_participants = 3
        num_sessions = 3

        # 3 sessions, 3 participants each is good to test alignment
        for i in range(num_sessions):
            otree.session.create_session(
                session_config_name=session_config_name,
                num_participants=num_participants,
            )

        with StringIO() as f:
            otree.export.export_wide(f)
            csv_text = f.getvalue()

        rows = [row for row in csv_text.split('\n') if row.strip()]

        # 1 row for each player + header row
        self.assertEqual(len(rows), num_participants * num_sessions + 1)

        header_row = rows[0]

        for expected_text in [
            'participant.id_in_session',
            'tests.export.1.player.id_in_group',
            'tests.export.1.group.group_field',
            'tests.export.1.subsession.subsession_field',
            'tests.export.2.subsession.subsession_field',
        ]:
            self.assertIn('{}'.format(expected_text), header_row)

        # 3.14 should be in the CSV without any currency formatting
        for expected_text in ['should be in export CSV', ',3.14,']:
            self.assertIn(expected_text, csv_text)

        # True/False should be converted to 1/0
        self.assertNotIn('False', csv_text)

        # test alignment
        for row in rows[1:]:
            participant_code = re.search(',ALIGN_TO_PARTICIPANT_(\w+),', row).group(1)
            self.assertIn(r',{},'.format(participant_code), row)

            # both rounds should have it
            self.assertEqual(
                row.count(',ALIGN_TO_PARTICIPANT_{}'.format(participant_code)),
                Constants.num_rounds,
            )

            session_code = re.search(',ALIGN_TO_SESSION_(\w+),', row).group(1)
            self.assertIn(',{},'.format(session_code), row)

            group_id = re.search(',GROUP_(\d+),', row).group(1)
            self.assertIn(',ALIGN_TO_GROUP_{},'.format(group_id), row)

            subsession_id = re.search(',SUBSESSION_(\d+),', row).group(1)
            self.assertIn(',ALIGN_TO_SUBSESSION_{},'.format(subsession_id), row)


from tests.export.pages import MyPage


class TestTimeSpentExport(TestCase):
    def setUp(self):
        self.client.login()

    def test_time_spent(self):
        page = MyPage()
        page.round_number = 1
        page.timeout_happened = True
        assert not PageTimeBatch.objects.exists()
        num_rows = BUFFER_SIZE + 10
        for i in range(num_rows):
            page._index_in_pages = i
            make_page_completion_row(
                view=page,
                app_name='app_name',
                participant__id_in_session=i,
                participant__code='part1c1pant',
                session_code='s3ss10n',
                is_wait_page=True,
            )
        # make sure batches are written periodically and not just on demand
        assert PageTimeBatch.objects.exists()

        response = self.client.get("/ExportPageTimes/")

        content_type = response["content-type"]
        assert content_type == "text/csv"

        content = response.content.decode('utf8')
        # make sure all content from the buffer is written on demand
        assert 's3ss10n,39,part1c1pant,39,app_name,MyPage' in content
        assert content.count('\n') >= num_rows + 1
