import otree.forms
from otree.api import widgets
from otree.db import models
from django.db import models as dj_models
from django.forms import widgets as dj_widgets
from .utils import TestCase
from .models import CurrencyFieldTestModel, SimpleModel, FormFieldModel


class SimpleModelForm(otree.forms.ModelForm):
    class Meta:
        model = FormFieldModel
        exclude = ()


class WidgetArgumentTests(TestCase):
    def test_widget_argument(self):
        self.assertIsInstance(
            SimpleModelForm.base_fields['char'].widget, dj_widgets.TextInput
        )
        self.assertIsInstance(
            SimpleModelForm.base_fields['text'].widget, dj_widgets.Textarea
        )
        self.assertIsInstance(
            SimpleModelForm.base_fields['alt_text'].widget, dj_widgets.Textarea
        )


class ModelTests(TestCase):
    def test_get_FIELD_display(self):
        obj = SimpleModel(name='bob')
        self.assertEqual(obj.get_name_display(), 'BOB')


class BooleanFieldTests(TestCase):
    def test_normal_widget_is_required_depending_on_blank(self):

        modelfield = models.BooleanField()
        formfield = modelfield.formfield()

        self.assertTrue(formfield.required)

        modelfield = models.BooleanField(blank=True)
        formfield = modelfield.formfield()

        self.assertFalse(formfield.required)

    def test_checkbox_form_field_is_not_required(self):

        modelfield = models.BooleanField(widget=widgets.CheckboxInput())
        formfield = modelfield.formfield()

        self.assertFalse(formfield.required)

    def test_checkbox_form_field_is_required_if_blank_is_set(self):

        # blank=True means en empty checkbox is ok.

        modelfield = models.BooleanField(widget=widgets.CheckboxInput(), blank=True)
        formfield = modelfield.formfield()

        self.assertFalse(formfield.required)

        # blank=False means the checkbox must be checked (is required).

        modelfield = models.BooleanField(widget=widgets.CheckboxInput(), blank=False)
        formfield = modelfield.formfield()

        self.assertTrue(formfield.required)


class CurrencyFieldTests(TestCase):
    def test_default_value_zero_in_modelform(self):
        # Make sure that a value of 0.00 is included in rendered output. It is
        # a falsy value if it's a Decimal instance. So this might fail if the
        # value is not converted to string prior to rendering in the form. See
        # https://github.com/oTree-org/otree-core/issues/326

        class CurrencyFieldTestModelForm(otree.forms.ModelForm):
            class Meta:
                model = CurrencyFieldTestModel
                exclude = ()

        form = CurrencyFieldTestModelForm()
        rendered = str(form['currency_with_default_value_zero'])
        self.assertTrue('value="0.00"' in rendered)


class FieldTests(TestCase):
    def test_label(self):
        '''label should be a replacement for verbose_name'''
        LABEL = 'yo'
        for FieldClass in [
            models.BooleanField,
            models.StringField,
            models.CurrencyField,
            models.IntegerField,
        ]:
            fld = FieldClass(label=LABEL)
            self.assertEqual(fld.verbose_name, LABEL)

            fld = FieldClass(verbose_name=LABEL)
            self.assertEqual(fld.verbose_name, LABEL)

    def test_initial(self):
        '''initial should be a replacement for default'''
        for FieldClass, initial_val in [
            (models.BooleanField, True),
            (models.StringField, 'yo'),
            (models.CurrencyField, 1.5),
            (models.IntegerField, 23),
        ]:
            fld = FieldClass(initial=initial_val)
            self.assertEqual(fld.default, initial_val)

            fld = FieldClass(default=initial_val)
            self.assertEqual(fld.default, initial_val)

    def test_stringfield(self):
        fld = models.StringField()
        # it seems to be django.db.models.fields.NOT_PROVIDED
        # self.assertEqual(fld.default, None)
        self.assertIsInstance(fld, dj_models.CharField)
        self.assertGreaterEqual(fld.max_length, 2000)
        formfield = fld.formfield()
        self.assertIsInstance(formfield.widget, dj_widgets.TextInput)

    def test_longstringfield(self):
        fld = models.LongStringField()
        # it seems to be django.db.models.fields.NOT_PROVIDED
        self.assertIsInstance(fld, dj_models.TextField)
        formfield = fld.formfield()
        self.assertIsInstance(formfield.widget, dj_widgets.Textarea)

    def test_null(self):
        for FieldClass in [
            models.BooleanField,
            models.StringField,
            models.CurrencyField,
            models.IntegerField,
            models.FloatField,
        ]:
            fld = FieldClass()
            self.assertEqual(fld.null, True)
