from unittest.mock import patch
from django.template import Context
from django.template import Template
from django.template import TemplateSyntaxError
from django.template import VariableDoesNotExist
from otree.session import create_session
import otree.db.models
import otree.forms

from .utils import TestCase
from .models import SimplePlayer
from .simple.models import Player


class PlayerForm(otree.forms.ModelForm):
    class Meta:
        model = Player
        fields = ('my_field',)


class SimplePlayerForm(otree.forms.ModelForm):
    class Meta:
        model = SimplePlayer
        fields = ('name', 'age')


class FormFieldTestMixin:
    def setUp(self):
        with patch('sys.stdout.write'):
            create_session('simple', num_participants=1)
        self.player = Player.objects.first()
        self.simple_player = SimplePlayer.objects.create()

    def parse(self, fragment):
        return Template('{% load otree %}' + fragment)

    def render(self, fragment, context=None):
        if context is None:
            context = Context()
        if not isinstance(context, Context):
            context = Context(context)
        return self.parse(fragment).render(context)


class FormFieldSyntaxTests(FormFieldTestMixin, TestCase):
    def test_fail_for_no_arguments(self):
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield %}')

    def test_fail_when_multiple_fields_are_given(self):
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield field field2 %}')

    def test_fail_with_bad_arguments_after_with(self):
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield field with label %}')
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield field -=- %}')
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield field label="Right" wrong %}')

    def test_with_syntax(self):
        self.parse('{% formfield field label="Hello" %}')
        self.parse('{% formfield field label=variable %}')
        self.parse('{% formfield field with label="Hello" %}')

    def test_one_field_is_given(self):
        self.parse('{% formfield field %}')
        self.parse('{% formfield player.field %}')
        self.parse('{% formfield player.soso.deeply.nested.field %}')


class FormFieldTests(FormFieldTestMixin, TestCase):
    def test_bound_field_displayed(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render('{% formfield form.name %}', {'form': form})
        self.assertInHTML(
            '''
            <input type="text" id="id_name" name="name" class="form-control"
                maxlength="50">
            ''',
            rendered,
        )
        # Label is there.
        self.assertInHTML(
            '''
            <label for="id_name" class="col-form-label">Name:</label>
            ''',
            rendered,
        )

    def test_model_field_displayed(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render(
            '{% formfield player.name %}', {'form': form, 'player': self.simple_player}
        )
        self.assertInHTML(
            '''
            <input type="text" id="id_name" name="name" class="form-control"
                maxlength="50">
            ''',
            rendered,
        )
        # Label is there.
        self.assertInHTML(
            '''
            <label for="id_name" class="col-form-label">Name:</label>
            ''',
            rendered,
        )

    def test_prefilled_model_field_displayed(self):
        player = SimplePlayer(name='Jack')
        form = SimplePlayerForm(instance=player)
        rendered = self.render(
            '{% formfield player.name %}', {'form': form, 'player': player}
        )
        self.assertInHTML(
            '''
            <input type="text" id="id_name" name="name" class="form-control"
                maxlength="50" value="Jack">
            ''',
            rendered,
        )

    def test_string_arg(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render("{% formfield 'name' %}", {'form': form})
        self.assertInHTML(
            '''
            <input type="text" id="id_name" name="name" class="form-control"
                maxlength="50">
            ''',
            rendered,
        )

    def test_string_arg_variable(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render("{% formfield myvar %}", {'form': form, 'myvar': 'name'})
        self.assertInHTML(
            '''
            <input type="text" id="id_name" name="name" class="form-control"
                maxlength="50">
            ''',
            rendered,
        )

    def test_model_field_not_part_of_form(self):
        form = SimplePlayerForm(instance=self.simple_player)

        # Different model instance of the same type.
        with self.assertRaises(ValueError) as cm:
            self.render(
                '{% formfield player.unkown_field %}',
                {'form': form, 'player': self.simple_player},
            )
        self.assertTrue("'unkown_field'" in str(cm.exception))

    def test_override_label(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render(
            '{% formfield player.name with label="A fancy label" %}',
            {'form': form, 'player': self.simple_player},
        )
        # Label is there.
        self.assertInHTML(
            '''
            <label for="id_name" class="col-form-label">A fancy label:</label>
            ''',
            rendered,
        )

    def test_empty_label_in_bootstrap_theme(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render(
            '''
            {% formfield player.name with label="" %}
            ''',
            {'form': form, 'player': self.simple_player},
        )
        # Label is there.
        self.assertTrue('<label' not in rendered, rendered)
