from unittest.mock import patch, MagicMock
from otree.models_concrete import CompletedGroupWaitPage
from django.db.backends.sqlite3.base import SQLiteCursorWrapper
from sqlite3 import OperationalError as SQLite3OperationalError
from otree.apps import (
    UndefinedColumn,
    UndefinedTable,
    patched_execute,
    SQLITE_LOCKING_ADVICE,
)
import django.db.utils
import pytest


@pytest.mark.django_db()
def test_integration():
    # i tried patching db.connection.cursor().execute
    # but cursor() returns a different cursor on each call.
    with patch.object(SQLiteCursorWrapper, 'execute') as execute:
        execute.side_effect = SQLite3OperationalError('database is locked!')
        # it seems Django converts it to utils.OperationalError
        with pytest.raises(django.db.utils.OperationalError) as excinfo:
            # make any random query
            CompletedGroupWaitPage.objects.exists()
        assert SQLITE_LOCKING_ADVICE in str(excinfo.value)


def get_fake_cursor(exc_to_raise):
    class FakeCursor:
        _execute = None  # because patched_execute references it
        execute = patched_execute

        def _execute_with_wrappers(self, *args, **kwargs):
            '''patched_execute calls this'''
            try:
                raise exc_to_raise
            except Exception as exc:
                # this mimics how the django does it.
                # UndefinedColumn is caught and re-raised as something else.
                # we do this so that __cause__ is populated.
                raise exc_to_raise from exc

    return FakeCursor()


def test_postgres_missing_table():

    with pytest.raises(UndefinedTable) as excinfo:
        get_fake_cursor(UndefinedTable).execute('')
    assert 'try resetting' in str(excinfo.value)


def test_sqlite_non_locking_issue():

    with pytest.raises(SQLite3OperationalError) as excinfo:
        get_fake_cursor(SQLite3OperationalError('dont use the L word')).execute('')
    assert SQLITE_LOCKING_ADVICE not in str(excinfo.value)


def test_other_error():
    '''make sure other exceptions don't get swallowed etc'''
    ERROR_MSG = 'blah blah!!'

    with pytest.raises(ZeroDivisionError) as excinfo:
        get_fake_cursor(ZeroDivisionError(ERROR_MSG)).execute('')
    assert str(excinfo.value) == ERROR_MSG
