from datetime import datetime
from unittest import mock

import pytest
from django.contrib.auth.models import User
from django.core.management import call_command
from django.test import override_settings, LiveServerTestCase
from django.urls import reverse

import otree.channels.utils as channel_utils
from otree.models import Session, Participant
from otree.session import create_session
from otree.views.mturk import MTurkCreateHIT
from tests.utils import OTreeChromeBrowser
from .utils import TestCase, AsyncTestCase, async_test, get_ws_communicator


class TestCreateSession(AsyncTestCase):
    @async_test
    async def test_create(self):
        br = self.splinter_browser
        br.login()
        br.visit_reverse('Sessions')
        br.click_link_by_partial_href('is_mturk=1')
        self.assertIn('Number of MTurk workers', br.html)
        self.assertIn('times more participant objects than', br.html)
        is_mturk_value = 'True'
        is_mturk_name = 'is_mturk'
        self.assertEqual(br.find_by_name(is_mturk_name).value, is_mturk_value)
        form_values = {
            'session_config': 'mturk',
            'num_participants': '1',
            is_mturk_name: is_mturk_value,
        }

        ws_client = get_ws_communicator(path=channel_utils.create_session_path())
        ws_client.scope['user'] = User.objects.first()

        await ws_client.connect()

        await ws_client.send_json_to(form_values)
        message_dict = await ws_client.receive_json_from()
        self.assertIn('MTurkCreateHIT', message_dict['session_url'])


class TestMTurk(TestCase):
    def setUp(self):
        create_session('mturk', num_participants=2, is_mturk=True)
        self.session: Session = Session.objects.get()
        self.session.save()
        self.splinter_browser.login()

    @mock.patch.object(MTurkCreateHIT, 'aws_keys_exist', True)
    @mock.patch.object(MTurkCreateHIT, 'boto3_installed', True)
    def test_get_create_hit_not_ready(self):
        # simulate using the old format
        del self.session.config['mturk_hit_settings']['template']
        self.session.save()
        br = self.splinter_browser
        br.visit_reverse('MTurkCreateHIT', args=[self.session.code])
        self.assertIn('MTurk is currently disabled', br.html)

    @mock.patch.object(MTurkCreateHIT, 'aws_keys_exist', True)
    @mock.patch.object(MTurkCreateHIT, 'boto3_installed', True)
    @override_settings(REAL_WORLD_CURRENCY_CODE='USD')
    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_create_hit(self, mock_get_mturk_client):

        mock_mturk_client = mock_get_mturk_client.return_value

        expiration = datetime(2070, 1, 1)
        create_hit = mock_mturk_client.create_hit

        # i guess this is because our own code depends on this,
        # so we need to mock it
        create_hit.return_value = {
            'HIT': {'HITId': 'AAA', 'HITGroupId': 'BBB', 'Expiration': expiration}
        }

        br = self.splinter_browser
        br.visit_reverse('MTurkCreateHIT', args=[self.session.code])
        br.uncheck('use_sandbox')
        button = br.find_by_id('btn-publish-hit')
        button.click()

        create_hit.assert_called()
        # make sure the right stuff gets sent to mturk
        # just a subset, things that are easy to mock
        expected_call_dict_subset = {
            'Title': 'Title for your experiment',
            'Description': 'Description for your experiment',
            'Keywords': 'easy, bonus, choice, study',
            'MaxAssignments': 1,
            'Reward': '0.4',
            'AssignmentDurationInSeconds': 3600,
            'LifetimeInSeconds': 60 * 60 * 4,
        }
        actual_call_dict = create_hit.call_args[1]
        for k, v in expected_call_dict_subset.items():
            assert actual_call_dict[k] == v

        # it's not there in sandbox mode
        self.assertIn('QualificationRequirements', actual_call_dict)

        session = Session.objects.get()

        self.assertIn('You have published', br.html)

        EXPIRED_TEXT = 'This HIT has expired'
        # let HIT expire by time
        with mock.patch('time.time', return_value=session.mturk_expiration + 1):
            br.reload()
            self.assertIn(EXPIRED_TEXT, br.html)
        br.reload()
        self.assertNotIn(EXPIRED_TEXT, br.html)
        button = br.find_by_id('btn-expire-hit')
        button.click()
        self.assertIn(EXPIRED_TEXT, br.html)

    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_mturk_start(self, mock_get_mturk_client):
        session = self.session
        # qualification requirements are not assigned in sandbox
        session.mturk_use_sandbox = False
        session.save()

        mocked_client = mock_get_mturk_client.return_value

        url = reverse('MTurkStart', args=(session.code,))
        # needs to start with 'A'
        worker_id = 'AWORKERID1'
        assignment_id = 'ASSIGNMENTID1'
        response = self.client.get(
            url,
            data={'assignmentId': assignment_id, 'workerId': worker_id},
            follow=True,
        )

        assert 200 == response.status_code

        # worker returns the assignment, accepts another assignment
        assignment_id2 = 'ASSIGNMENTID2'
        response = self.client.get(
            url,
            data={'assignmentId': assignment_id2, 'workerId': worker_id},
            follow=True,
        )

        mocked_client.associate_qualification_with_worker.assert_called()

        # should be only 1
        p_visitor = Participant.objects.get(session=session, visited=True)
        p_non_visitor = Participant.objects.filter(session=session, visited=False)[0]

        self.assertEqual(p_visitor.mturk_worker_id, worker_id)
        self.assertEqual(p_visitor.mturk_assignment_id, assignment_id2)
        self.assertEqual(p_non_visitor.mturk_worker_id, None)
        self.assertEqual(p_non_visitor.mturk_assignment_id, None)

    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_mturk_sandbox(self, mock_get_mturk_client):
        session = self.session
        # qualification requirements are not assigned in sandbox
        session.mturk_use_sandbox = True
        session.save()

        mocked_client = mock_get_mturk_client.return_value

        url = reverse('MTurkStart', args=(session.code,))
        # needs to start with 'A'
        worker_id = 'AWORKERID1'
        assignment_id = 'ASSIGNMENTID1'
        response = self.client.get(
            url,
            data={'assignmentId': assignment_id, 'workerId': worker_id},
            follow=True,
        )

        mocked_client.associate_qualification_with_worker.assert_not_called()


@pytest.mark.slow
class TestPayMTurkE2E(LiveServerTestCase):

    '''
    Use LiveServerTestCase instead of ChannelLiveServerTestCase, because
    I need to use mocks, which are not possible with ChannelLiveServerTestCase
    because it uses a separate process
    '''

    def setUp(self):
        num_participants = 6
        # create a session
        call_command('create_session', 'mturk', str(num_participants))
        self.session = Session.objects.get()
        self.session.mturk_HITId = 'FAKEHITID'
        self.session.save()
        self.participants = list(self.session.get_participants())
        self.browser = OTreeChromeBrowser(live_server_url=self.live_server_url)
        self.browser.login()

    # so we don't have to worry about points conversion.
    @override_settings(USE_POINTS=False)
    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_pay(self, mock_get_mturk_client):

        mocked_client = mock_get_mturk_client()

        bonus_per_participant = 0.2
        self.assertEqual(self.session.config['participation_fee'], 0.4)

        # simulate workers arriving from MTurk by assigning bogus worker IDs
        # and assignment IDs
        for p in self.participants:
            p.mturk_assignment_id = p.mturk_worker_id = str(p.id_in_session)
            p.payoff = bonus_per_participant
            p.save()

        # mock the result set, so that when the view queries MTurk to see
        # who is submitted, it returns a list of MTurk workers

        def list_assignments_for_hit(**kwargs):
            # don't paginate
            if 'NextToken' in kwargs:
                return {'Assignments': []}
            return {
                'NextToken': 'foo',
                'Assignments': [
                    {
                        'WorkerId': p.mturk_worker_id,
                        'AssignmentId': p.mturk_worker_id,
                        'AssignmentStatus': 'Submitted',
                        'Answer': '',
                    }
                    # simulate that not everyone submitted
                    for p in self.participants[:-2]
                ],
            }

        mocked_client.list_assignments_for_hit = list_assignments_for_hit

        url = reverse('MTurkSessionPayments', args=[self.session.code])
        br = self.browser
        br.go(url)

        br.find_by_css('[name="workers"][value="1"]').check()
        br.find_by_css('[name="workers"][value="2"]').check()

        button = br.find_by_css('button#pay')

        self.assertEqual(button.text, 'Pay via MTurk ($1.20)')
        button.click()

        confirm_button = br.find_by_id('pay-confirm-button')
        confirm_button.click()

        for asst_id in ['1', '2']:
            mocked_client.approve_assignment.assert_any_call(AssignmentId=asst_id)

        calls = mocked_client.send_bonus.call_args_list
        self.assertEqual(len(calls), 2)

        for call in calls:
            kwargs = call[1]
            self.assertEqual(kwargs['BonusAmount'], '0.20')
            # because our bogus worker ID is the same as our bogus assignment ID
            self.assertEqual(kwargs['WorkerId'], kwargs['AssignmentId'])

        # after submitting, oTree redirects to the payments page
        br.find_by_css('[name="workers"][value="3"]').check()
        br.find_by_css('[name="workers"][value="4"]').check()

        button = br.find_by_css('button#reject')
        button.click()
        reject_confirm_button = br.find_by_id('reject-confirm-button')
        reject_confirm_button.click()

        calls = mocked_client.reject_assignment.call_args_list
        self.assertEqual(len(calls), 2)
        for asst_id in ['3', '4']:
            mocked_client.reject_assignment.assert_any_call(
                AssignmentId=asst_id, RequesterFeedback=''
            )


class TestPayMTurk(TestCase):
    def setUp(self):
        num_participants = 6
        # create a session
        call_command('create_session', 'mturk', str(num_participants))
        self.session = Session.objects.get()
        self.session.mturk_HITId = 'FAKEHITID'
        self.session.save()
        self.participants = list(self.session.get_participants())

    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_auto_reject(self, mock_get_mturk_client):

        mocked_client = mock_get_mturk_client()

        def list_assignments_for_hit(**kwargs):
            # don't paginate
            if 'NextToken' in kwargs:
                return {'Assignments': []}
            return {
                'NextToken': 'foo',
                'Assignments': [
                    {
                        'WorkerId': 'fake',
                        'AssignmentId': 'fake',
                        'AssignmentStatus': 'Submitted',
                        'Answer': '',
                    }
                ],
            }

        mocked_client.list_assignments_for_hit = list_assignments_for_hit

        url = reverse('MTurkSessionPayments', args=[self.session.code])
        cl = self.client
        cl.login()
        cl.get(url)

        mocked_client.reject_assignment.assert_called()
