import os
import shutil
import subprocess
import sys
import tempfile
from os.path import join
import pytest
from django.core.management import call_command

from otree.management.commands.resetdb import (
    logger as resetdb_logger,
    MSG_DB_ENGINE_FOR_HUB,
    MSG_RESETDB_SUCCESS_FOR_HUB,
)
from unittest.mock import patch
from .utils import TestCase, cd


class ResetDBMockTest(TestCase):
    @patch('otree.management.commands.resetdb.cursor_execute_drop_cmd', autospec=True)
    @patch('otree.management.commands.resetdb.migrate_db', autospec=True)
    def test_resetdb(self, patched_migrate_db, cursor_drop_cmd):
        with self.assertLogs(resetdb_logger) as cm:
            call_command('resetdb', '--noinput')
            logger_output = '\n'.join(cm.output)
            # these are required for oTree Hub to function
            # todo: I think this assert is somehow sensitive to test order
            # because it broke around the time i installed the 'order tests randomly'
            # plugin
            self.assertIn(MSG_DB_ENGINE_FOR_HUB, logger_output)
            self.assertIn(MSG_RESETDB_SUCCESS_FOR_HUB, logger_output)
        patched_migrate_db.assert_called()
        cursor_drop_cmd.assert_called()


@pytest.mark.slow
class ResetDBTest(TestCase):
    '''Test startproject, startapp, resetdb'''

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def test_startproject_resetdb(self):
        tmp_dir = self.tmp_dir
        project_name = 'newproj'
        args = f'otree startproject {project_name} --noinput'.split()
        env = os.environ.copy()
        del env['DJANGO_SETTINGS_MODULE']

        # better to do this than call_command, because it's more realistic
        # for example, call_command runs when settings are already loaded.
        # this caused an actual issue on 2018-09-30 with INSTALLED_APPS.
        process = subprocess.Popen(
            args, cwd=tmp_dir, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = process.communicate()
        self.assertIn(b'Created project folder', out)

        proj_path = os.path.join(tmp_dir, project_name)
        self.assertTrue(os.path.exists(proj_path))
        self.assertTrue(os.path.isdir(proj_path))
        settings_path = join(proj_path, 'settings.py')
        self.assertTrue(os.path.exists(settings_path))

        with cd(proj_path), patch('sys.stdout.write'):
            call_command('startapp', 'foo')
            with open(settings_path, 'a') as settings_file:
                # simpler than putting it in SESSION_CONFIGS
                config = {
                    'name': 'foo',
                    'display_name': 'foo',
                    'app_sequence': ['foo'],
                    'num_demo_participants': 1,
                }
                settings_file.write('\n')
                settings_file.write('SESSION_CONFIGS.append({})'.format(config))

        # now test resetdb.
        # need to use a separate process because don't want to tamper with
        # the test database
        args = [sys.executable, 'manage.py', 'resetdb', '--noinput']
        env = os.environ.copy()
        del env['DJANGO_SETTINGS_MODULE']
        process = subprocess.Popen(
            args, cwd=proj_path, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = process.communicate()
        self.assertIn(b'Creating table foo_player', out)
        self.assertIn(b'Created new tables and columns', err)
