from .utils import TestCase, get_path
from json import dumps, loads
from otree.models import Session
from django.conf import settings
from unittest import mock
import os
from django.shortcuts import reverse
from otree.session import create_session


def add_label(base_url, label):
    return base_url + '?participant_label={}'.format(label)


from .utils import REST_KEY

SESSIONS_API_URL = '/api/sessions/'
PARTICIPANT_VARS_API_URL = '/api/participant_vars/'


class ParticipantVarsTests(TestCase):
    def setUp(self):

        self.cl2 = self.client_class(HTTP_OTREE_REST_KEY=REST_KEY)
        self.room_name = settings.ROOM_WITHOUT_LABELS_NAME
        self.charlie = 'Charlie'

    def participant_visits(self):
        ROOM_URL = reverse(
            'AssignVisitorToRoom', args=[settings.ROOM_WITHOUT_LABELS_NAME]
        )
        start_url = add_label(ROOM_URL, self.charlie)
        self.cl2.get(start_url, follow=True)

    def post_vars(self):
        self.vars_to_set = dict(pvar_1=314, pvar_2=3.14, pvar_3=[1, 2, [5], {'x': 1}])

        payload = dict(
            room_name=self.room_name,
            participant_label=self.charlie,
            vars=self.vars_to_set,
        )

        resp = self.cl2.post(
            PARTICIPANT_VARS_API_URL, dumps(payload), content_type="application/json"
        )
        assert resp.status_code == 200

    def post_session(self):
        payload = dict(
            session_config_name='rest_api', num_participants=2, room_name=self.room_name
        )
        resp = self.cl2.post(
            SESSIONS_API_URL, dumps(payload), content_type="application/json"
        )
        return resp.content.decode()

    def run_asserts(self, session_code):
        p1, p2 = Session.objects.get(code=session_code).get_participants()
        assert p1.label == self.charlie
        assert p1.vars == self.vars_to_set
        assert p2.label == None
        assert p2.vars == {}

    @mock.patch.dict(os.environ, {'OTREE_REST_KEY': REST_KEY})
    def test_vars_before_participant(self):
        session_code = self.post_session()
        self.post_vars()
        self.participant_visits()
        self.run_asserts(session_code)

    @mock.patch.dict(os.environ, {'OTREE_REST_KEY': REST_KEY})
    def test_participant_before_vars(self):
        session_code = self.post_session()
        self.participant_visits()
        self.post_vars()
        self.run_asserts(session_code)


class SessionVarsTests(TestCase):
    @mock.patch.dict(os.environ, {'OTREE_REST_KEY': REST_KEY})
    def test_vars(self):
        room_name = settings.ROOM_WITHOUT_LABELS_NAME

        session = create_session(
            session_config_name='rest_api', num_participants=1, room_name=room_name
        )
        session.vars = dict(a=1, b=False)
        session.save()

        payload = dict(room_name=room_name, vars=dict(b=True, c=[1, 2]))

        client = self.client_class(HTTP_OTREE_REST_KEY=REST_KEY)
        resp = client.post(
            '/api/session_vars/', dumps(payload), content_type="application/json"
        )
        assert resp.status_code == 200, resp.content
        session.refresh_from_db()
        assert session.vars == dict(a=1, b=True, c=[1, 2])


class CreateSessionTests(TestCase):
    def setUp(self):
        self.cl2 = self.client_class(HTTP_OTREE_REST_KEY=REST_KEY)

    @mock.patch.dict(os.environ, {'OTREE_REST_KEY': REST_KEY})
    def test_basic(self):
        cl = self.cl2

        payload = dict(session_config_name='rest_api', num_participants=4)
        resp = cl.post(
            SESSIONS_API_URL, dumps(payload), content_type="application/json"
        )
        assert resp.status_code == 200
        session_code = resp.content.decode('utf8')
        assert Session.objects.filter(code=session_code).exists()

    @mock.patch.dict(os.environ, {'OTREE_REST_KEY': REST_KEY})
    def test_no_auth(self):
        cl = self.cl2

        payload = dict(session_config_name='rest_api', num_participants=4)
        # use self.client, which doesn't have the special header.
        resp = self.client.post(
            SESSIONS_API_URL, dumps(payload), content_type="application/json"
        )
        assert resp.status_code == 403

    @mock.patch.dict(os.environ, {'OTREE_REST_KEY': REST_KEY})
    def test_wrong_auth(self):
        cl = self.client_class(HTTP_OTREE_REST_KEY='wrong_key')

        payload = dict(session_config_name='rest_api', num_participants=4)
        # use self.client, which doesn't have the special header.
        resp = cl.post(
            SESSIONS_API_URL, dumps(payload), content_type="application/json"
        )
        assert resp.status_code == 403

    @mock.patch.dict(os.environ, {'OTREE_REST_KEY': REST_KEY})
    def test_advanced(self):
        ROOM_NAME = settings.ROOM_WITH_LABELS_NAME
        cl = self.cl2
        STR_VALUE = 'added_through_api'
        modified_session_config_fields = dict(
            int=7,
            float=0.7,
            list=[7, 7],
            str='edited_through_api',
            participation_fee=7.77,
            a_added=STR_VALUE,
        )

        payload = dict(
            session_config_name='rest_api',
            room_name=ROOM_NAME,
            num_participants=4,
            modified_session_config_fields=modified_session_config_fields,
        )
        resp = cl.post(
            SESSIONS_API_URL, dumps(payload), content_type="application/json"
        )
        assert resp.status_code == 200
        session_code = resp.content.decode('utf8')
        session = Session.objects.get(code=session_code)
        assert session.config['a_added'] == STR_VALUE  # sanity check
        assert modified_session_config_fields.items() <= session.config.items()
        room = session.get_room()
        assert room.name == ROOM_NAME
