from unittest.mock import patch

import django.test
import otree.channels.utils as channel_utils
import splinter
from django.conf import settings
from django.urls import reverse
from otree.models.participant import Participant
from otree.session import create_session
from tests.utils import ConnectingWSClient, OTreeSplinterDjangoBrowser
from .utils import (
    TestCase,
    get_app_communicator,
    get_ws_communicator,
    execute_mocked_send_to_channel,
    AsyncTestCase,
    AsyncMock,
    async_test,
    extract_message_sent_to_mock,
    execute_mocked_send_group,
)
from otree.common import add_params_to_url
from django.contrib.auth.models import User


from .utils import TestCase

URL_ADMIN_LABELS = reverse('RoomWithoutSession', args=[settings.ROOM_WITH_LABELS_NAME])
URL_ADMIN_NO_LABELS = reverse(
    'RoomWithoutSession', args=[settings.ROOM_WITHOUT_LABELS_NAME]
)
URL_ADMIN_SECURE_URLS = reverse(
    'RoomWithoutSession', args=[settings.ROOM_WITH_SECURE_URLS_NAME]
)
URL_PARTICIPANT_LABELS = reverse(
    'AssignVisitorToRoom', args=[settings.ROOM_WITH_LABELS_NAME]
)
URL_PARTICIPANT_NO_LABELS = reverse(
    'AssignVisitorToRoom', args=[settings.ROOM_WITHOUT_LABELS_NAME]
)
URL_PARTICIPANT_SECURE_URLS = reverse(
    'AssignVisitorToRoom', args=[settings.ROOM_WITH_SECURE_URLS_NAME]
)

LABEL_REAL = 'JohnSmith'
LABEL_REAL2 = 'Bob'
LABEL_FAKE = 'NotInParticipantLabelsFile'

WAITING_STR = 'Waiting for your session to begin'


def add_label(base_url, label):
    return base_url + '?participant_label={}'.format(label)


class RoomTestCase(TestCase):
    def setUp(self):
        # should explicitly use admin or participant browser
        self._splinter_browser_disabled = True
        self.participant_browser = OTreeSplinterDjangoBrowser()
        self.admin_browser = OTreeSplinterDjangoBrowser()
        self.admin_browser.login()


class SecureURLTests(RoomTestCase):
    def test_click_link(self):
        # visiting without hash code or participant label should give 404
        # visiting with participant label
        br = self.admin_browser

        br.visit(URL_ADMIN_SECURE_URLS)
        br.click_link_by_partial_href(LABEL_REAL)
        self.assertIn(WAITING_STR, br.html)

    def test_missing_hash(self):
        br = self.participant_browser
        url_without_hash = add_params_to_url(
            URL_PARTICIPANT_SECURE_URLS, {'participant_label': LABEL_REAL}
        )

        br.visit(url_without_hash)
        self.assertEqual(br.status_code, 404)

    def test_wrong_hash(self):
        url_wrong_hash = add_params_to_url(
            URL_PARTICIPANT_SECURE_URLS,
            {'participant_label': LABEL_REAL, 'hash': 'fakehash'},
        )

        br = self.participant_browser
        br.visit(url_wrong_hash)
        self.assertEqual(br.status_code, 404)

    def test_has_session_but_missing_hash(self):

        create_session(
            'simple',
            # make it 6 so that we can test if the participant is reassigned
            # if they open their start link again (1/6 chance)
            num_participants=1,
            room_name=settings.ROOM_WITH_SECURE_URLS_NAME,
        )

        url_without_hash = add_params_to_url(
            URL_PARTICIPANT_SECURE_URLS, {'participant_label': LABEL_REAL}
        )

        br = self.participant_browser
        br.visit(url_without_hash)
        self.assertEqual(br.status_code, 404)


class TestRoomWithoutSession(RoomTestCase):
    def test_open_admin_links(self):
        urls = [reverse('Rooms'), URL_ADMIN_LABELS, URL_ADMIN_NO_LABELS]

        br = self.admin_browser

        for url in urls:
            br.visit(url)

    def test_open_participant_links(self):

        br = self.participant_browser

        br.visit(URL_PARTICIPANT_LABELS)
        self.assertIn('Please enter your participant label', br.html)

        br.visit(add_label(URL_PARTICIPANT_LABELS, LABEL_REAL))

        self.assertIn(WAITING_STR, br.html)

        br.visit(add_label(URL_PARTICIPANT_LABELS, LABEL_FAKE))
        self.assertIn('Invalid entry', br.html)

        br.visit(URL_PARTICIPANT_NO_LABELS)
        self.assertIn(WAITING_STR, br.html)

        br.visit(add_label(URL_PARTICIPANT_NO_LABELS, LABEL_REAL))
        self.assertIn(WAITING_STR, br.html)

    def test_ping(self):
        pass


class TestRoomWithSession(RoomTestCase):
    def _visited_count(self, session):
        return session.participant_set.filter(visited=True).count()

    def visited_count_labels(self):
        return self._visited_count(self.session_with_labels)

    def visited_count_no_labels(self):
        return self._visited_count(self.session_without_labels)

    def setUp(self):
        super().setUp()
        self.session_with_labels = create_session(
            'simple',
            # make it 6 so that we can test if the participant is reassigned
            # if they open their start link again (1/6 chance)
            num_participants=6,
            room_name=settings.ROOM_WITH_LABELS_NAME,
        )

        self.session_without_labels = create_session(
            'simple', num_participants=2, room_name=settings.ROOM_WITHOUT_LABELS_NAME
        )

    def tearDown(self):
        br = self.admin_browser
        for url in [URL_ADMIN_LABELS, URL_ADMIN_NO_LABELS]:
            br.visit(url)
            button = br.find_by_id('close-room')
            button.click()

        br.visit(URL_ADMIN_LABELS)
        self.assertIn('room_without_session', br.url)

    def test_open_admin_links(self):
        br = self.admin_browser
        br.visit(URL_ADMIN_LABELS)
        self.assertIn('room_with_session', br.url)

    def test_without_label(self):
        br = self.participant_browser

        br.visit(URL_PARTICIPANT_NO_LABELS)
        br.visit(URL_PARTICIPANT_NO_LABELS)

        # should use a cookie, so it remembers the participant
        self.assertEqual(self.visited_count_no_labels(), 1)

        br.visit(add_label(URL_PARTICIPANT_NO_LABELS, LABEL_REAL))
        self.assertEqual(self.visited_count_no_labels(), 2)

    def test_with_label(self):
        br = self.participant_browser

        br.visit(URL_PARTICIPANT_LABELS)
        self.assertIn('Please enter your participant label', br.html)

        br.visit(add_label(URL_PARTICIPANT_LABELS, LABEL_FAKE))
        self.assertIn('Invalid entry', br.html)
        self.assertEqual(self.visited_count_labels(), 0)

        # make sure reopening the same link assigns you to same participant
        for i in range(2):
            br.visit(add_label(URL_PARTICIPANT_LABELS, LABEL_REAL))

        visited_participants = Participant.objects.filter(visited=True)
        self.assertEqual(len(visited_participants), 1)

        participant = visited_participants[0]

        self.assertEqual(participant.label, LABEL_REAL)

    def test_delete_session_in_room(self):
        pass

    def test_session_start_links_room(self):
        pass


class PresenceWithLabelsTests(AsyncTestCase):
    def setUp(self):
        super().setUp()
        self.participant_label = LABEL_REAL
        room_name = settings.ROOM_WITH_LABELS_NAME
        self.room_name = room_name
        self.tab_unique_id = '12123123'
        self.path = channel_utils.room_participant_path(
            room_name=room_name,
            participant_label=self.participant_label,
            tab_unique_id=self.tab_unique_id,
        )
        self.admin_path = channel_utils.room_admin_path(room_name)
        self.participant_ws_client = get_ws_communicator(self.path)
        self.admin_ws_client = get_ws_communicator(self.admin_path)
        self.admin_ws_client.scope['user'] = User.objects.first()

    @patch('time.time')
    @async_test
    async def test_stale_visits(self, patched_time):

        patched_time.return_value = 0

        # participant connects
        await self.participant_ws_client.connect()
        await self.participant_ws_client.receive_nothing()

        # time passes, participant does not disconnect but becomes stale
        patched_time.return_value = 30

        # admin requests StaleRoomVisits, should include that participant

        admin_http_client = self.client_class()
        admin_http_client.login()
        heartbeat_url = reverse('ParticipantRoomHeartbeat', args=[self.tab_unique_id])
        stale_room_visits_url = reverse('StaleRoomVisits', args=[self.room_name])

        resp = admin_http_client.get(stale_room_visits_url)
        self.assertJSONEqual(
            resp.content.decode('utf8'),
            {'participant_labels': [self.participant_label]},
        )

        # participant heartbeat occurs
        django.test.Client().get(heartbeat_url)

        resp = admin_http_client.get(stale_room_visits_url)
        self.assertJSONEqual(resp.content.decode('utf8'), {'participant_labels': []})

    @async_test
    async def test_participant_before_admin(self):

        participant_client = self.participant_ws_client
        admin_client = self.admin_ws_client

        await participant_client.connect()
        await participant_client.receive_nothing()

        await admin_client.connect()

        # check that DB record created properly
        self.assertEqual(
            await admin_client.receive_json_from(),
            {
                'status': 'load_participant_lists',
                'participants_present': [self.participant_label],
            },
        )

    @async_test
    async def test_admin_before_participant(self):
        participant_client = self.participant_ws_client
        admin_client = self.admin_ws_client

        await admin_client.connect()
        self.assertEqual(
            await admin_client.receive_json_from(),
            {'status': 'load_participant_lists', 'participants_present': []},
        )

        await participant_client.connect()
        await participant_client.receive_nothing()

        data = await admin_client.receive_json_from()
        self.assertEqual(
            data, {'status': 'add_participant', 'participant': self.participant_label}
        )

    @async_test
    async def test_participant_disconnect_w(self):
        participant_client = self.participant_ws_client
        admin_client = self.admin_ws_client

        # can only call .wait() once?

        await participant_client.connect()
        # i need to either call receive_json_from or receive_nothing!
        # that's how to force it to do something!
        await participant_client.receive_nothing()

        # i don't understand why the .disconnect() method did not cause
        # the .disconnect() consumer to be run.

        await admin_client.connect()
        await admin_client.receive_json_from()

        await participant_client.disconnect()
        await participant_client.wait()

        self.assertEqual(
            await admin_client.receive_json_from(),
            {'status': 'remove_participant', 'participant': self.participant_label},
        )

    @async_test
    async def test_with_session(self):
        participant_client = self.participant_ws_client
        create_session(
            'simple', num_participants=2, room_name=settings.ROOM_WITH_LABELS_NAME
        )
        await participant_client.connect()
        self.assertEqual(
            await participant_client.receive_json_from(), {'status': 'session_ready'}
        )


class PresenceWithoutLabelsTests(AsyncTestCase):
    def setUp(self):
        super().setUp()
        room_name = settings.ROOM_WITHOUT_LABELS_NAME
        self.room_name = room_name
        self.tab_unique_id = '12123123'
        self.path = channel_utils.room_participant_path(
            room_name=room_name, participant_label='', tab_unique_id=self.tab_unique_id
        )
        self.admin_path = channel_utils.room_admin_path(room_name)
        self.participant_ws_client = get_ws_communicator(self.path)
        self.admin_ws_client = get_ws_communicator(self.admin_path)
        self.admin_ws_client.scope['user'] = User.objects.first()

    @patch('time.time')
    @async_test
    async def test_stale_visits(self, patched_time):

        patched_time.return_value = 0

        # participant connects
        await self.participant_ws_client.connect()
        await self.participant_ws_client.receive_nothing()

        # time passes, participant does not disconnect but becomes stale
        patched_time.return_value = 30

        # admin requests StaleRoomVisits, should include that participant
        admin_http_client = self.client
        admin_http_client.login()
        heartbeat_url = reverse('ParticipantRoomHeartbeat', args=[self.tab_unique_id])
        participant_count_url = reverse(
            'ActiveRoomParticipantsCount', args=[self.room_name]
        )

        resp = admin_http_client.get(participant_count_url)
        self.assertJSONEqual(resp.content.decode('utf8'), {'count': 0})

        # participant heartbeat occurs
        django.test.Client().get(heartbeat_url)

        resp = admin_http_client.get(participant_count_url)
        self.assertJSONEqual(resp.content.decode('utf8'), {'count': 1})

    @async_test
    async def test_participant_before_admin_wo(self):

        participant_client = self.participant_ws_client
        admin_client = self.admin_ws_client

        await participant_client.connect()
        await participant_client.receive_nothing()

        await admin_client.connect()

        # check that DB record created properly
        self.assertEqual(
            await admin_client.receive_json_from(),
            {
                'status': 'load_participant_lists',
                # a list with length 1, but no label
                'participants_present': [''],
            },
        )

    @async_test
    async def test_admin_before_participant(self):
        participant_client = self.participant_ws_client
        admin_client = self.admin_ws_client

        await admin_client.connect()
        self.assertEqual(
            await admin_client.receive_json_from(),
            {'status': 'load_participant_lists', 'participants_present': []},
        )

        await participant_client.connect()
        await participant_client.receive_nothing()

        self.assertEqual(
            await admin_client.receive_json_from(),
            {'status': 'add_participant', 'participant': ''},
        )

    @async_test
    async def test_participant_disconnect(self):
        participant_client = self.participant_ws_client
        admin_client = self.admin_ws_client

        await participant_client.connect()
        await participant_client.receive_nothing()

        await admin_client.connect()
        await admin_client.receive_json_from()

        await participant_client.disconnect(timeout=2)
        await participant_client.wait()

        self.assertEqual(
            await admin_client.receive_json_from(), {'status': 'remove_participant'}
        )


# issues:
# - ChannelLiveServerTestCase can not be used with in memory databases, so i need to somehow choose a different DB.
# - Weird error "Can't pickle DaphneProcess": https://github.com/django/channels/issues/1207. no resolution.
'''
from channels.testing import ChannelsLiveServerTestCase
from .utils import OTreeChromeBrowser

class TestRoomJS(ChannelsLiveServerTestCase):

    def test_presence(self):
        # admin browser
        abr = OTreeChromeBrowser(live_server_url=self.live_server_url)

        # participant browser
        pbr = OTreeChromeBrowser(live_server_url=self.live_server_url)

        # participant opens waiting page
        pbr.go(URL_PARTICIPANT_LABELS)

        pbr.fill('participant_label', LABEL_REAL)
        button = pbr.find_by_tag('button')
        button.click()
        waiting_url = pbr.url

        # admin opens RoomWithoutSession
        abr.go(URL_ADMIN_LABELS)

        # within a few seconds, he should see that participant online,
        # but not another participant.

        presence_indicator_selector = '#present-participant-label-{}.count-this'
        john_presence_selector = presence_indicator_selector.format(LABEL_REAL)
        bob_presence_selector = presence_indicator_selector.format(LABEL_REAL2)

        self.assertTrue(
            abr.is_element_present_by_css(john_presence_selector, wait_time=5))
        self.assertFalse(
            abr.is_element_present_by_css(bob_presence_selector))

        # if participant closes the browser, he should go offline
        # go to some other URL
        pbr.go('/')

        self.assertFalse(
            abr.is_element_present_by_css(john_presence_selector, wait_time=1))

        # comes back online
        # need to use visit, not go, because it's an absolute URL
        pbr.visit(waiting_url)

        # admin creates a session in the room
        abr.fill_form({
            'session_config': 'simple',
            'num_participants': '1',
        })
        button = abr.find_by_value('Create')
        button.click()

        # within a few seconds, the participant should be redirected
        my_field_present = pbr.is_element_present_by_name('my_field', wait_time=2)
        self.assertTrue(my_field_present)
'''
