from unittest import mock
from unittest.mock import Mock

import django.db
import django.db.models

from otree.api import BaseConstants
from otree.bots.runner import run_bots
from otree.currency import Currency
from otree.db import models
from otree.db import idmap
from otree.db.vars import VarsError
from otree.models import Participant, Session
from otree.session import create_session
from tests.simple.models import Player
from .utils import TestCase


class SavingTests(TestCase):
    # We need to make sure to flush the idmap cache here after every save in
    # order to prevent getting values that do not actually represent the DB
    # values.

    def test_dont_save_if_no_change(self):
        create_session('simple', num_participants=1)

        with idmap.use_cache():
            participant = Participant.objects.first()

            # We keep track of the participant.
            instances = list(idmap.get_instances())
            self.assertEqual(instances, [participant])

            # But we won't save the participant since we didn't change it.
            with self.assertNumQueries(0):
                idmap.save()

    def test_save_only_changed_fields(self):
        create_session('simple', num_participants=1)
        with idmap.use_cache():
            participant = Participant.objects.first()
            with mock.patch.object(django.db.models.Model, 'save') as patched_save:
                idmap.save()
                # in save-the-change 2017 version, save is not called at all
                patched_save.assert_called_once_with(update_fields=[])

                participant.code = 'hello'
                idmap.save()
                # but is this necessarily the only argument? no
                patched_save.assert_called_with(update_fields=['code'])

    def test_nested_changes(self):
        create_session('simple', num_participants=1)
        # Reset cache.
        with idmap.use_cache():
            # Query participant via session.
            session = Session.objects.first()
            participant = session.participant_set.first()
            participant.is_on_wait_page = not participant.is_on_wait_page

            # Save participant.
            with self.assertNumQueries(1):
                idmap.save()

    def test_with_app_models(self):
        create_session('simple', num_participants=2)

        from .simple.models import Player

        with idmap.use_cache():
            players = Player.objects.all()
            self.assertEqual(len(players), 2)

            group = players[0].group
            group.save = Mock()
            group.round_number += 1

            # Query session object to test that it's loaded..
            group.session
            participants = group.session.participant_set.all()

            all_instances = {
                players[0],
                players[1],
                group,
                group.session,
                participants[0],
                participants[1],
            }

            self.assertEqual(set(idmap.get_instances()), all_instances)

            # No queries are executed. The group model shall be saved, but we
            # mocked out the save method. All other models should be left
            # untouched.
            # 2020-10-08: my new field tracker doesnt guarantee this.
            # it seems that participants have .session_id set on them.
            # with self.assertNumQueries(0):
            idmap.save()

            group.save.assert_called()


class SavingBots(TestCase):
    def test_bots(self):
        session = create_session('saving', num_participants=1)
        run_bots(session)


class PickleFieldTests(TestCase):
    def test_pickle(self):
        test_values = [
            {"a": [], "b": None, "c": [1, 3.0, {}], 'd': {Currency(1): Currency(2)}}
        ]
        for value in test_values:
            field = models._PickleField(default=value)
            serialized = field.get_prep_value(value)
            restored = field.to_python(serialized)
            self.assertEqual(value, restored)


class Constants(BaseConstants):
    alist = [1, 2]


class VarsTests(TestCase):
    def test_vars_are_saved(self):
        o = Session.objects.create()
        o.vars['a'] = 'b'
        o.save()

        o = Session.objects.get()
        self.assertEqual(o.vars, {'a': 'b'})

        o.vars['a'] = 'd'
        o.save()

        o = Session.objects.get()
        self.assertEqual(o.vars, {'a': 'd'})

    def test_forbid_storing_models(self):
        create_session('simple', num_participants=1)

        session = Session.objects.get()
        participant = Participant.objects.get()
        player = Player.objects.get()

        for value in [player, [player], {1: player}, participant]:
            session.vars['something_with_a_model'] = value
            with self.assertRaises(VarsError):
                session.save()

    def test_storing_constants(self):

        self.assertEqual(Constants.alist, [1, 2])
        o = Session.objects.create()
        # users should always be using .copy(), even when assigning to vars,
        # because they might modify it before it's saved.
        o.vars['alist'] = Constants.alist.copy()
        o.save()

        o = Session.objects.get()
        vars_alist = o.vars['alist']
        self.assertEqual(vars_alist, Constants.alist)
        vars_alist.append(3)

        self.assertEqual(vars_alist, [1, 2, 3])
        # should not modify the original
        self.assertEqual(Constants.alist, [1, 2])

    def test_numpy(self):
        '''
        Previously:
        https://github.com/karanlyons/django-save-the-change/issues/27

        The bug occurs when the dict keys are the same,
        so it recurses to check equality of the values,
        and one of the values is a numpy array

        This issue reoccurred when I switched to model_utils FieldTracker.

        '''

        o = Session.objects.create()
        o.vars['array'] = FakeNumpyArray()
        o.save()

        o = Session.objects.get()
        array = o.vars['array']

        # test that it was round-tripped without data loss
        self.assertIsInstance(array, FakeNumpyArray)
        o.save()


class FakeNumpyArray:
    def __ne__(self, other):
        return self

    def __bool__(self):
        raise ValueError(
            'The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()'
        )
