import pytest
from django.template import Context, Template

from otree.bots.bot import ParticipantBot
from otree.session import create_session
from otree_startup.settings import InvalidVariableError
from .utils import TestCase


class TestTemplates(TestCase):
    def test_all_blocks_found(self):
        session = create_session(
            session_config_name='templates_app', num_participants=1
        )

        participant = session.participant_set.first()
        bot = ParticipantBot(participant, player_bots=[])
        bot.open_start_url()
        for block_name in [
            'scripts',
            'app_scripts',
            'global_scripts',
            'styles',
            'app_styles',
            'global_styles',
        ]:
            self.assertIn('my_{}'.format(block_name), bot.html)


class TestMissingValues(TestCase):
    def evaluate(self, fragment: str, **kwargs):
        return Template(fragment).render(Context(kwargs))

    def good(self, fragment: str, **kwargs):
        self.evaluate(fragment, **kwargs)

    def bad(self, fragment: str, **kwargs):
        with self.assertRaises(InvalidVariableError, msg=fragment):
            self.evaluate(fragment, **kwargs)

    def test_missing_values(self):
        class Player:
            pass

        player = Player()

        d = {'real': True}

        bad = self.bad
        good = self.good

        bad('{{ bogus }}')
        bad('{{ player.bogus }}', player=player)
        bad('{{ d.fake }}', d=d)
        bad('{% if bogus %}{% endif %}')
        bad('{% if player.bogus %}{% endif %}', player=player)
        bad('{% if player.bogus == 1 %}{% endif %}', player=player)
        bad('{% if bogus != 1 %}{% endif %}')

        good('{% if False %} {{bogus}} {% endif %}')
        good('{{ d.real }}', d=d)
        good('{{ d|default:None }}')
        good('{{ d.e.f|default:None }}')
        good('{{ player.bogus|default:None }}', player=player)
        good('{{ d.fake|default:None }}', d=d)
        good('{% if bogus|default:None == 1 %}{% endif %}')


@pytest.mark.django_db(transaction=True)
def test_javascript_vars():
    session = create_session(session_config_name='javascript_vars', num_participants=2)

    participant = session.participant_set.first()
    bot = ParticipantBot(participant, player_bots=[])
    bot.open_start_url()

    expected = '<script>var js_vars = {"a": 1.0, "b": null};</script>'
    # can't use .html because we are looking for stuff that's before otree-content div
    actual = bot._response.content.decode('utf8')
    assert expected in actual

    bot.submit(post_data={})
    expected = '<script>var js_vars = {"jsList": ["a", 42.0]};</script>'
    actual = bot._response.content.decode('utf8')
    assert expected in actual
