from unittest import mock
from unittest.mock import patch
import pytest
import tests.wait_page.pages
import tests.wait_page.pages
import tests.waitpage_skip_race.pages
import tests.waitpage_template.models
from otree.bots.bot import ParticipantBot
from otree.bots.runner import run_bots
from otree.views import abstract
from otree.session import create_session
from .utils import (
    TestCase,
    get_ws_communicator,
    AsyncTestCase,
    async_test,
    execute_mocked_send_group,
)


class TestWaitForAllGroups(TestCase):
    def setUp(self):
        session = create_session('wait_page', num_participants=4)
        subsession = session.get_subsessions()[0]
        self.group1 = subsession.get_groups()[0]

    def start_1_group(self, should_be_stuck):
        bots = []
        for player in self.group1.get_players():
            bot = ParticipantBot(player.participant, player_bots=[])
            bots.append(bot)
        for bot in bots:
            bot.open_start_url()
        for bot in bots:
            bot.open_start_url()
            self.assertEqual(bot.on_wait_page(), should_be_stuck)

    def test_dont_wait_for_all(self):
        self.start_1_group(should_be_stuck=False)

    def test_wait_for_all_groups(self):
        with mock.patch.object(
            tests.wait_page.pages.MyWait,
            'wait_for_all_groups',
            new_callable=mock.PropertyMock,
            return_value=True,
        ):
            self.start_1_group(should_be_stuck=True)


class TestSkipWaitPage(TestCase):
    def setUp(self):
        session = create_session('skip_wait_page', num_participants=2)
        bots = []
        for participant in session.get_participants():
            bot = ParticipantBot(participant, player_bots=[])
            bots.append(bot)
        self.bots = bots

    def visit(self, ordered_bots):
        for bot in ordered_bots:
            bot.open_start_url()
        for bot in ordered_bots:
            bot.open_start_url()
            self.assertFalse(bot.on_wait_page())

    def test_skipper_visits_last(self):
        self.visit(self.bots)

    def test_waiter_visits_last(self):
        self.visit(reversed(self.bots))


class TemplateTests(TestCase):
    def test_customization(self):

        # need 2 participants so the wait page is not skipped
        session = create_session('waitpage_template', num_participants=2)
        participant = session.participant_set.first()
        br = self.splinter_browser
        Constants = tests.waitpage_template.models.Constants
        with self.assertTemplateUsed(template_name=Constants.wait_page_template):
            br.visit(participant._start_url())
        for substring in [Constants.custom_title_text, Constants.custom_body_text]:
            self.assertTrue(br.is_text_present(substring))


class Wrapper:
    '''Wrapper for base class so that test runner doesn't run this'''

    class RaceTestsBase(AsyncTestCase):

        '''Race conditions in connecting to wait pages'''

        config_name: str
        WaitPageClass = None

        def play(self, participant):
            self.splinter_browser.visit(participant._start_url())

        def get_ws_client(self):
            page = self.WaitPageClass()
            self.p1.refresh_from_db()
            page.set_attributes(participant=self.p1)

            ws_client = get_ws_communicator(path=page.socket_url())
            return ws_client

        def setUp(self):
            super().setUp()
            session = create_session(self.config_name, num_participants=2)
            self.p1, self.p2 = session.get_participants()

        @patch('otree.channels.utils._sync_group_send')
        @async_test
        async def test_slow_websocket(self, mock_a2s):
            p1 = self.p1
            p2 = self.p2

            self.play(p1)
            self.play(p2)

            await execute_mocked_send_group(mock_a2s)

            p1_client = self.get_ws_client()
            await p1_client.connect()
            self.assertEqual(await p1_client.receive_json_from(), {'status': 'ready'})

        @patch('otree.channels.utils._sync_group_send')
        @async_test
        async def test_other_player_last(self, mock_a2s):
            p1 = self.p1
            p2 = self.p2

            self.play(p1)

            p1_client = self.get_ws_client()
            await p1_client.connect()

            self.play(p2)
            await execute_mocked_send_group(mock_a2s)

            self.assertEqual(await p1_client.receive_json_from(), {'status': 'ready'})


class RaceTests(Wrapper.RaceTestsBase):
    config_name = 'wait_page'
    WaitPageClass = tests.wait_page.pages.MyWait


class SkipRaceTests(Wrapper.RaceTestsBase):

    '''Need a separate test for wait pages that are skipped, because
    they go through a different code path (_increment_index_in_pages)
    '''

    config_name = 'waitpage_skip_race'
    WaitPageClass = tests.waitpage_skip_race.pages.MyWait

    def play(self, participant):
        br = self.splinter_browser
        br.visit(participant._start_url())
        br.find_by_tag('button').first.click()


@pytest.mark.django_db()
def test_aapa_string():
    session = create_session(
        session_config_name='waitpage_aapa_string', num_participants=4
    )

    run_bots(session)
    session.refresh_from_db()
    assert session.vars['aapa_subsession'] == True
    [subsession] = session.get_subsessions()
    g1, g2 = subsession.get_groups()
    for g in [g1, g2]:
        assert g.f1 == True
        assert g.f2 == True
