# -*- coding: utf-8 -*-

from django.utils import translation
from django.utils.encoding import force_text

from django import forms
import otree.db.models
import otree.forms
from otree.api import widgets  # new style import
from .utils import TestCase
from otree.api import BasePlayer, models, widgets
from otree.currency import Currency


class BasicWidgetTests(TestCase):
    def test_attrs_yield_form_control_class(self):
        class Form(otree.forms.Form):
            first_name = otree.forms.CharField()

        rendered = force_text(Form()['first_name'])
        elem = (
            '<input type="text" name="first_name" '
            'required class="form-control" id="id_first_name" />'
        )
        self.assertHTMLEqual(rendered, elem)


class CurrencyInputTests(TestCase):
    maxDiff = None

    class CurrencyForm(forms.Form):
        currency = otree.forms.CurrencyField()

    def test_widget(self):
        form = self.CurrencyForm({'currency': Currency('52.23')})
        rendered = force_text(form['currency'])
        self.assertInHTML(u'''<span class="input-group-text">€</span>''', rendered)
        # Make sure that the value does not contain the currency symbol
        self.assertIn('''value="52.23"''', rendered)


class RadioSelectHorizontalTests(TestCase):
    maxDiff = None

    class RadioForm(forms.Form):
        numbers = forms.ChoiceField(
            choices=(('1', '1'), ('2', '2')), widget=otree.forms.RadioSelectHorizontal
        )

    def test_widget(self):
        form = self.RadioForm()

        rendered = force_text(form['numbers'])
        self.assertHTMLEqual(
            rendered,
            """
            <div class="form-check form-check-inline">
                <input type="radio" name="numbers" value="1" required id="id_numbers_0" class="form-check-input" />
                <label for="id_numbers_0" class="form-check-label">1</label>
            </div>

            <div class="form-check form-check-inline">
                <input type="radio" name="numbers" value="2" required id="id_numbers_1" class="form-check-input" />
                <label for="id_numbers_1" class="form-check-label">2</label>
            </div>
            """,
        )


class RadioSelectTests(TestCase):
    '''
    good to have a test case for this, because it relies on
    input_option.html which we have customized.
    '''

    maxDiff = None

    class RadioForm(forms.Form):
        numbers = forms.ChoiceField(
            choices=(('1', 'one'),), widget=otree.forms.RadioSelect
        )

    def test_widget(self):
        form = self.RadioForm()

        rendered = force_text(form['numbers'])
        self.assertHTMLEqual(
            rendered,
            """
            <ul id="id_numbers">
              <li>
                <label for="id_numbers_0">
                  <input type="radio" name="numbers" value="1"
                         required id="id_numbers_0" class="form-check-input">
                  one
                </label>
              </li>
            </ul>
            """,
        )


class CheckboxInputTests(TestCase):
    maxDiff = None

    class CheckboxInputTestsModel(otree.db.models.OTreeModel):
        booleanfield = models.BooleanField(widget=widgets.CheckboxInput)

    def test_with_booleanfield(self):
        class Form(otree.forms.ModelForm):
            class Meta:
                model = self.CheckboxInputTestsModel
                fields = ('booleanfield',)

        form = Form(data={'booleanfield': 'on'})
        assert form.is_valid()
        assert form.cleaned_data['booleanfield'] is True

        form = Form(data={})
        assert form.is_valid()
        assert form.cleaned_data['booleanfield'] is False


class RadioInputTests(TestCase):
    def test_radio_no_blank(self):

        # Fields with a RadioSelect should be rendered without the
        # '---------' option, and with nothing selected by default, to
        # see otree.forms.forms.BaseModelForm.__init__

        class RadioModel(otree.db.models.OTreeModel):
            f_radio = models.IntegerField(
                blank=True, choices=[1, 2], widget=widgets.RadioSelect
            )

        class Form(otree.forms.ModelForm):
            class Meta:
                model = RadioModel
                fields = ('f_radio',)

        form = Form()
        radio_text = force_text(form['f_radio'])
        self.assertNotIn('--------', radio_text)
        self.assertNotIn('value=""', radio_text)


class SliderTests(TestCase):
    maxDiff = None

    class SlideInputTestsModel(otree.db.models.OTreeModel):

        currencyfield = models.CurrencyField(
            min=0,
            max=100,
            initial=98,
            # legacy name is SliderInput
            widget=widgets.SliderInput(attrs={'step': Currency(5)}),
        )
        floatfield = models.FloatField(
            min=0, max=2.5, default=0, widget=widgets.Slider(attrs={'step': '0.01'})
        )

    def test_sliderinput_with_float_field(self):
        class Form(otree.forms.ModelForm):
            class Meta:
                model = self.SlideInputTestsModel
                fields = ('floatfield',)

        form = Form()
        self.assertHTMLEqual(
            force_text(form['floatfield']),
            '''
            <div class="input-group slider" data-slider>
                <input
                    type="range"
                    name="floatfield"
                    value="0"
                    required
                    max="2.5"
                    step="0.01"
                    id="id_floatfield"
                    min="0"
                    class="form-control"
                >
                <div class="input-group-append">
                    <span class="input-group-text" 
                    data-slider-value title="current value"></span>
                </div>
            </div>
            ''',
        )

    def test_sliderinput_with_float_field_and_locale_set(self):
        with translation.override('de-de'):
            self.test_sliderinput_with_float_field()

    def test_with_currencyfield(self):
        class Form(otree.forms.ModelForm):
            class Meta:
                model = self.SlideInputTestsModel
                fields = ('currencyfield',)

        form = Form()
        self.assertHTMLEqual(
            force_text(form['currencyfield']),
            '''
            <div class="input-group slider" data-slider>
                <input
                    type="range"
                    name="currencyfield"
                    value="98"
                    required
                    max="100"
                    step="5.00"
                    id="id_currencyfield"
                    min="0"
                    class="form-control"
                >
                <div class="input-group-append">
                    <span class="input-group-text" 
                    data-slider-value title="current value"></span>
                </div>
            </div>
            ''',
        )

    def test_with_currencyfield_and_locale(self):
        with translation.override('de-de'):
            self.test_with_currencyfield()
