import shutil
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch, MagicMock, create_autospec, call

from django.core.management import call_command

from otree.management.commands.zip import (
    validate_reqs_files,
    fix_reqs_files,
    RequirementsError,
    zip_project,
    OVERWRITE_TOKEN,
    DONT_OVERWRITE_TOKEN,
)
from otree_startup import execute_from_command_line
from otree_startup import zipserver
from .utils import TestCase, cd


class RequirementsTests(TestCase):
    def test_validate(self):
        test_files = Path(__file__).parent / 'zip'

        good = test_files / 'good'
        bad = test_files / 'bad'

        for path in good.iterdir():
            validate_reqs_files(path)

        for path in bad.iterdir():
            with self.assertRaises(RequirementsError, msg=path):
                validate_reqs_files(path)

    def test_fix(self):
        test_files = Path(__file__).parent.joinpath('zip/fix')

        with patch('builtins.input') as patched_input:
            assert fix_reqs_files(test_files / 'ignore') == {}
            patched_input.assert_not_called()

        with patch('builtins.input') as patched_input:
            patched_input.return_value = 'n'
            d = fix_reqs_files(test_files / 'legacy')
            assert DONT_OVERWRITE_TOKEN in d['requirements.txt']

        with patch('builtins.input') as patched_input:
            patched_input.return_value = 'y'
            d = fix_reqs_files(test_files / 'legacy')
            assert OVERWRITE_TOKEN in d['requirements.txt']
            assert 'otree>=' in d['requirements.txt']

        with patch('builtins.input') as patched_input:
            patched_input.return_value = 'y'
            d = fix_reqs_files(test_files / 'can_overwrite_no_base')
            assert 'requirements_base' not in d

        with patch('builtins.input') as patched_input:
            patched_input.return_value = 'y'
            d = fix_reqs_files(test_files / 'can_overwrite_mturk')
            assert 'otree[mturk]>=' in d['requirements.txt']


class ZipTests(TestCase):
    @patch('builtins.input')
    def test_zip(self, patched_input):
        with TemporaryDirectory() as root:
            root_path = Path(root)
            project_path = root_path / 'project'

            shutil.copytree('zip/test_project', str(project_path))
            with cd(project_path):
                with patch(
                    'otree.management.commands.zip.validate_reqs_files'
                ) as patched_validate:
                    patched_input.return_value = 'y'
                    zip_project(project_path)
                    patched_input.assert_called()
                    patched_validate.assert_called()

            # can't unzip when you are already in a project
            shutil.move(str(project_path / 'project.otreezip'), root_path)

            with cd(root_path):
                # call this rather than unzip function for fuller coverage
                call_command('unzip', 'project.otreezip')

            # it should automatically be named project-2
            new_project_path = root_path / 'project-2'
            self.assertTrue((new_project_path / 'requirements.txt').exists())
            self.assertFalse((new_project_path / 'db.sqlite3').exists())


class RunzipTests(TestCase):
    @patch('otree_startup.zipserver.sleep')
    @patch('otree_startup.zipserver.stdout_write')
    def test_zipserver_before_otreezip(self, patched_stdout: MagicMock, *args):
        fake_project = create_autospec(zipserver.Project, instance=True)

        fake_project.mtime.return_value = 1.0
        fake_project.poll.side_effect = [None, 1]
        FAKE_PROJECT_NAME = 'foo.otreezip'
        fake_project.zipname.return_value = FAKE_PROJECT_NAME

        with patch(
            'otree_startup.zipserver.get_newest_project',
            # need to keep returning fake_project until poll returns non-null
            side_effect=[None, None, fake_project, fake_project],
        ):
            zipserver.autoreload_for_new_zipfiles()

            patched_stdout.assert_has_calls(
                [
                    call(zipserver.MSG_NO_OTREEZIP_YET),
                    call(zipserver.MSG_RUNNING_OTREEZIP_NAME.format(FAKE_PROJECT_NAME)),
                ]
            )

    @patch('otree_startup.zipserver.sleep')
    @patch('otree_startup.zipserver.stdout_write')
    def test_change_otreezip(self, patched_stdout: MagicMock, *args):
        p1 = create_autospec(zipserver.Project, instance=True)
        p1.mtime.return_value = 1.0
        p1.poll.return_value = None

        p2 = create_autospec(zipserver.Project, instance=True)
        p2.mtime.return_value = 2.0
        p2.poll.side_effect = [None, 1]
        p2.zipname.return_value = 'foo.otreezip'

        with patch(
            'otree_startup.zipserver.get_newest_project',
            # need to keep returning fake_project until poll returns non-null
            side_effect=[p1, p2, p2],
        ):
            zipserver.autoreload_for_new_zipfiles()
            p1.terminate.assert_called()
            patched_stdout.assert_has_calls(
                [
                    call(zipserver.MSG_FOUND_NEWER_OTREEZIP),
                    call(zipserver.MSG_RUNNING_OTREEZIP_NAME.format('foo.otreezip')),
                ]
            )

    @patch('otree_startup.zipserver.stdout_write')
    def test_get_newest_project(self, patched_stdout: MagicMock):
        with TemporaryDirectory() as td_name:
            with cd(td_name):
                for x in range(zipserver.MAX_OTREEZIP_FILES + 1):
                    Path(f'{x}.otreezip').touch()
                    # this seems to ensure different mtime
                    # without it, get flaky results
                    # 0.001 always worked on my PC, but not GitHub CI
                    time.sleep(0.01)

                p = zipserver.get_newest_project()
                # last value of x
                self.assertEqual(p.zipname(), f'{x}.otreezip')
                patched_stdout.assert_called_with(
                    zipserver.MSG_DELETING_OLD_OTREEZIP.format('0.otreezip')
                )

    @patch('sys.exit')
    @patch('otree_startup.argv', "otree zipserver xyz.otreezip".split())
    @patch('otree_startup.zipserver.Project', autospec=True)
    def test_cli_onefile(self, MockProject, *args):
        project = MockProject.return_value
        project.wait.return_value = 0
        execute_from_command_line()
        project.wait.assert_called()

    @patch('sys.exit')
    @patch('otree_startup.argv', "otree zipserver".split())
    @patch(
        'otree_startup.zipserver.autoreload_for_new_zipfiles', spec=True, return_value=0
    )
    def test_cli(self, autoreload_fn, *args):
        execute_from_command_line()
        autoreload_fn.assert_called()
