import contextlib
import os
import shutil
import sys
import tempfile
from unittest.mock import patch

from channels.testing import WebsocketCommunicator, ApplicationCommunicator

import otree.channels.utils as channel_utils
import asyncio
from unittest.mock import Mock

import django.test
import splinter.browser

import urllib

from django.conf import settings
from django.core.management import call_command
from django.test.utils import override_settings

from django.shortcuts import reverse

from django.contrib.auth.models import User

from otree.channels.routing import application

REST_KEY = 'restkey'


@contextlib.contextmanager
def add_path(path):
    """
    ::

        with add_path(new_sys_path):
            import strange_module
    """
    sys.path.insert(0, path)
    yield
    sys.path.pop(sys.path.index(path))


@contextlib.contextmanager
def cd(directory):
    """
    ::

        with cd(new_cwd):
            os.walk('.')
    """
    old_path = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(old_path)


@contextlib.contextmanager
def dummyapp(app):
    """
    Creates a new game app inside a temporary directory, adds this to the path
    and puts the app into the INSTALLED_APPS and INSTALLED_OTREE_APPS settings.

    It cleans the filesystem up afterwards.
    """
    tmpdir = tempfile.mkdtemp()
    app_path = os.path.join(tmpdir, app)

    os.mkdir(app_path)
    with patch('sys.stdout.write'):
        call_command('startapp', app, app_path)

    new_apps = list(settings.INSTALLED_APPS) + [app]
    new_otree_apps = list(settings.INSTALLED_OTREE_APPS) + [app]
    with add_path(tmpdir):
        with override_settings(
            INSTALLED_APPS=new_apps, INSTALLED_OTREE_APPS=new_otree_apps
        ):
            yield app_path

        shutil.rmtree(tmpdir)
        if os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)


def get_path(test_client_response, if_no_redirect):
    try:
        url = test_client_response.redirect_chain[-1][0]
    except IndexError:
        return if_no_redirect
    else:
        return urllib.parse.urlsplit(url).path


class OTreeTestClient(django.test.client.Client):
    def login(self):
        return super().force_login(user=User.objects.get())


# Switch to Chrome when I have a chance. For now, the mturk test is failing,
# and the process keeps running even after I press Ctrl+C
class OTreeChromeBrowser(splinter.browser.ChromeWebDriver):

    # splinter.Browser is actually a function, not a class
    def __init__(self, live_server_url):
        self.live_server_url = live_server_url
        # eventually, pass headless=True
        # but for now, it seems the browser is never killedgit
        super().__init__(headless=True)

    def go(self, relative_url):
        self.visit(self.live_server_url + relative_url)

    def login(self, next_url="/"):
        self.go(reverse('login') + '?next=' + next_url)
        self.fill_form(
            dict(username=settings.ADMIN_USERNAME, password=settings.ADMIN_PASSWORD)
        )
        self.find_by_id('btn-login').click()


class OTreeSplinterDjangoBrowser(splinter.browser.DjangoClient):
    def __init__(self, django_test_client: OTreeTestClient = None):
        super().__init__()
        # _browser is attribute Splinter uses to store the Django test client
        # we store our subclassed test client in there
        if not django_test_client:
            django_test_client = OTreeTestClient()
        self._browser = django_test_client

    def login(self):
        self._browser.login()

    def visit_reverse(self, viewname, args=None, kwargs=None):
        url = reverse(viewname, args=args, kwargs=kwargs)
        self.visit(url)


class ConnectingWSClient:
    pass


class _SplinterBrowserMixin:
    client_class = OTreeTestClient

    client: OTreeTestClient
    _splinter_browser = None
    _splinter_browser_disabled = False

    @property
    def splinter_browser(self) -> OTreeSplinterDjangoBrowser:
        if self._splinter_browser_disabled:
            raise AttributeError
        if not self._splinter_browser:
            self._splinter_browser = OTreeSplinterDjangoBrowser(self.client)
        return self._splinter_browser


class TestCase(_SplinterBrowserMixin, django.test.TestCase):
    pass


def async_test(coro):
    """
    Async test wrapper courtesy of: https://stackoverflow.com/a/23036785/3362468
    """

    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(coro(*args, **kwargs))

    return wrapper


def AsyncMock():
    """
    Substitute Mock for mocking/patching coroutines.  Courtesy of:
    https://stackoverflow.com/a/32505333/3362468
    """
    coro = Mock(name="CoroutineResult")
    corofunc = Mock(name="CoroutineFunction", side_effect=asyncio.coroutine(coro))
    corofunc.coro = coro
    return corofunc


from django.test import TransactionTestCase

# class AsyncTestCase(_SplinterBrowserMixin, django.test.TestCase):
class AsyncTestCase(_SplinterBrowserMixin, TransactionTestCase):
    '''
    Need to inherit from TransactionTestCase, otherwise I get this error in
    many test cases:
    django.db.utils.OperationalError: database table is locked
    If I use postgres
    I'm not sure where the transaction even is.
    This is slow.
    even an empty test case takes 300ms. according to my profiling
    it seems to be because of the
    transaction, not because of setup/teardown and async stuff,
    which takes less than 0.001 seconds.
    Using pytest isn't any faster; i need to use
    @pytest.mark.django_db(transaction=True)
    '''

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        tasks = asyncio.gather(
            *asyncio.all_tasks(loop=self.loop), loop=self.loop, return_exceptions=True
        )
        tasks.add_done_callback(lambda t: self.loop.stop())
        tasks.cancel()

        while not tasks.done() and not self.loop.is_closed():
            self.loop.run_forever()

        self.loop.close()


async def execute_mocked_send_group(mocked):
    '''
    https://github.com/django/channels/issues/944#issuecomment-500159157
    '''
    call_args_list = mocked.call_args_list
    if len(call_args_list) == 0:
        raise Exception(f'No calls were made to send_group')
    for args, kwargs in call_args_list:
        await channel_utils._group_send(*args, **kwargs)
    mocked.reset_mock()


async def execute_mocked_send_to_channel(mocked_send):
    '''
    https://github.com/django/channels/issues/944#issuecomment-500159157
    '''

    [call_args] = mocked_send.call_args_list
    [channel_name, message], _ = call_args
    comm = ApplicationCommunicator(
        application=application, scope={'type': 'channel', 'channel': channel_name}
    )
    await comm.send_input(message)
    await comm.wait()
    # return await comm.receive_output(timeout=4)


def extract_message_sent_to_mock(mock_send) -> dict:
    [channel_name, message], _ = mock_send.call_args
    return message


class OTreeWebsocketCommunicator(WebsocketCommunicator):
    '''
    async def connect_and_wait(self):
        await self.connect()
        await self.wait()
    '''


def get_ws_communicator(path: str, login=False) -> OTreeWebsocketCommunicator:
    '''login should be passed explictly, so that my tests accurately verify login requirements'''

    ws_client = OTreeWebsocketCommunicator(application=application, path=path)
    if login:
        ws_client.scope['user'] = User.objects.first()
    return ws_client


def get_app_communicator(channel: str) -> ApplicationCommunicator:
    return ApplicationCommunicator(
        application=application, scope={'type': 'channel', 'channel': channel}
    )
